from io import StringIO

from django.core.management import call_command

from app_utils.testing import NoSocketsTestCase

from memberaudit.tests.testdata.factories import create_character_contract
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import create_memberaudit_character


class TestResetCharacters(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    def test_should_reset_section_data_for_characters(self):
        # given
        character_1001 = create_memberaudit_character(1001)
        create_character_contract(character_1001)
        out = StringIO()

        # when
        call_command("memberaudit_reset_characters", "--noinput", stdout=out)

        # then
        self.assertFalse(character_1001.contracts.exists())



# Generated by Django 4.0.10 on 2023-12-13 23:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0016_extend_characterupdatestatus"),
    ]

    operations = [
        migrations.AlterField(
            model_name="characterasset",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
    ]



import json
import shutil
import tempfile
from unittest.mock import patch

from django.test import TestCase

from memberaudit.models import Character, _helpers
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.utils import create_memberaudit_character

MODULE_PATH = "memberaudit.models._helpers"


@patch(MODULE_PATH + ".settings")
class TestStoreCharacterData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def setUp(self) -> None:
        self.root_path = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.root_path)

    def test_should_store_when_enabled(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertTrue(result.exists())

        with result.open("r") as file:
            data_2 = json.load(file)

        self.assertEqual(data, data_2)

    def test_should_store_when_section_enabled(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True), patch(
            MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_SECTIONS", ["assets"]
        ):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertTrue(result.exists())

    def test_should_also_accept_section_obj(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section=Character.UpdateSection.ASSETS,
            )

        # then
        self.assertTrue(result.exists())

    def test_should_not_store_when_section_not_enabled(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True), patch(
            MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_SECTIONS", ["other_section"]
        ):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertIsNone(result)

    def test_should_store_when_character_enabled(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True), patch(
            MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS", [self.character.id]
        ):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertTrue(result.exists())

    def test_should_not_store_when_character_not_enabled(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True), patch(
            MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS", [-1]
        ):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertIsNone(result)

    def test_should_allow_optional_suffix(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
                suffix="special",
            )

        # then
        self.assertTrue(result.exists())
        self.assertIn("special", result.name)

    def test_should_not_break_when_section_name_is_wrong(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True):
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="invalid",
            )

        # then
        self.assertIsNone(result)

    def test_should_not_break_when_json_serialization_fails(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True), patch(
            MODULE_PATH + ".json.dump"
        ) as mock:
            mock.side_effect = TypeError
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertIsNone(result)

    def test_should_not_break_when_write_failed(self, mock_settings):
        # given
        mock_settings.BASE_DIR = self.root_path
        data = [{"dummy": 1}]

        # when
        with patch(MODULE_PATH + ".MEMBERAUDIT_STORE_ESI_DATA_ENABLED", True), patch(
            MODULE_PATH + ".json.dump"
        ) as mock:
            mock.side_effect = OSError
            result = _helpers.store_character_data_to_disk_when_enabled(
                character=self.character,
                data=data,
                section="assets",
            )

        # then
        self.assertIsNone(result)



"""Managers for character section models (2/3)."""
# pylint: disable=missing-class-docstring

from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any, Dict, List, Set

from bravado.exception import HTTPNotFound

from django.db import models, transaction
from esi.models import Token
from eveuniverse.models import (
    EveAncestry,
    EveBloodline,
    EveEntity,
    EveFaction,
    EveRace,
    EveSolarSystem,
    EveType,
)

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import (
    MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
    MEMBERAUDIT_MAX_MAILS,
)
from memberaudit.core.xml_converter import eve_xml_to_html
from memberaudit.decorators import fetch_token_for_character
from memberaudit.helpers import UpdateSectionResult, data_retention_cutoff
from memberaudit.models._helpers import store_character_data_to_disk_when_enabled
from memberaudit.providers import esi
from memberaudit.utils import (
    get_or_create_esi_or_none,
    get_or_create_or_none,
    get_or_none,
)

from ._common import GenericUpdateComplexObjMixin, GenericUpdateSimpleObjMixin

if TYPE_CHECKING:
    from memberaudit.models import Character, CharacterMail

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterCorporationHistoryManager(GenericUpdateComplexObjMixin, models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create corporation history for character."""

        return character.update_section_if_changed(
            section=character.UpdateSection.CORPORATION_HISTORY,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    def _fetch_data_from_esi(self, character: Character) -> List[dict]:
        logger.info("%s: Fetching corporation history from ESI", character)
        history = esi.client.Character.get_characters_character_id_corporationhistory(
            character_id=character.eve_character.character_id,
        ).results()

        return history

    def _update_or_create_objs(
        self, character: Character, esi_data: List[dict]
    ) -> Set[int]:
        def make_obj_from_esi_entry(character: Character, entry: dict):
            corporation = get_or_create_or_none("corporation_id", entry, EveEntity)
            obj = self.model(
                character=character,
                record_id=entry["record_id"],
                corporation=corporation,
                is_deleted=entry.get("is_deleted"),
                start_date=entry["start_date"],
            )
            return obj

        new_eve_entity_ids = self._update_or_create_objs_generic(
            character,
            esi_data,
            model_key_field="record_id",
            fields_for_update=("corporation_id", "is_deleted", "start_date"),
            make_obj_from_esi_entry=make_obj_from_esi_entry,
            return_new_eve_entities=True,
        )
        return new_eve_entity_ids


class CharacterDetailsManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create character details from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.CHARACTER_DETAILS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    def _fetch_data_from_esi(self, character: Character):
        logger.info("%s: Fetching character details from ESI", character)
        details = esi.client.Character.get_characters_character_id(
            character_id=character.eve_character.character_id,
        ).results()

        return details

    def _update_or_create_objs(self, character: Character, details) -> Set[int]:
        description = (
            details.get("description", "") if details.get("description") else ""
        )

        # TODO: remove when fixed
        # temporary fix to address u-bug in ESI endpoint for character bio
        # workaround to address syntax error bug (#77)
        # see also: https://github.com/esi/esi-issues/issues/1265
        if description and description.startswith("u'") and description.endswith("'"):
            try:
                description = ast.literal_eval(description)
            except SyntaxError:
                logger.warning("Failed to convert description with u-bug.")
                description = ""

        if description:
            eve_xml_to_html(description)  # resolve entities early
        gender = (
            self.model.GENDER_MALE
            if details.get("gender") == "male"
            else self.model.GENDER_FEMALE
        )

        # TODO: Remove once issue is fixed
        # Workaround because of ESI issue #1264
        eve_ancestry = get_or_none("ancestry_id", details, EveAncestry)

        obj, _ = self.update_or_create(
            character=character,
            defaults={
                "alliance": get_or_create_or_none("alliance_id", details, EveEntity),
                "birthday": details.get("birthday"),
                "eve_ancestry": eve_ancestry,
                "eve_bloodline": get_or_create_esi_or_none(
                    "bloodline_id", details, EveBloodline
                ),
                "eve_faction": get_or_create_esi_or_none(
                    "faction_id", details, EveFaction
                ),
                "eve_race": get_or_create_esi_or_none("race_id", details, EveRace),
                "corporation": get_or_create_or_none(
                    "corporation_id", details, EveEntity
                ),
                "description": description,
                "gender": gender,
                "name": details.get("name", ""),
                "security_status": details.get("security_status"),
                "title": details.get("title", "") if details.get("title") else "",
            },
        )
        return obj.eve_entity_ids()


class CharacterFwStatsManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create fw stats for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.FW_STATS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_fw_stats.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token):
        logger.info("%s: Fetching FW stats from ESI", character)
        stats = esi.client.Faction_Warfare.get_characters_character_id_fw_stats(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return stats

    def _update_or_create_objs(self, character: Character, stats: dict):
        if faction_id := stats.get("faction_id"):
            faction, _ = EveFaction.objects.get_or_create_esi(id=faction_id)
        else:
            faction = None
        self.update_or_create(
            character=character,
            defaults={
                "current_rank": stats.get("current_rank"),
                "enlisted_on": stats.get("enlisted_on"),
                "faction": faction,
                "highest_rank": stats.get("highest_rank"),
                "kills_last_week": stats["kills"]["last_week"],
                "kills_total": stats["kills"]["total"],
                "kills_yesterday": stats["kills"]["yesterday"],
                "victory_points_last_week": stats["victory_points"]["last_week"],
                "victory_points_total": stats["victory_points"]["total"],
                "victory_points_yesterday": stats["victory_points"]["yesterday"],
            },
        )


class CharacterImplantManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create implants for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.IMPLANTS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-clones.read_implants.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> List[int]:
        logger.info("%s: Fetching implants from ESI", character)
        implant_type_ids = esi.client.Clones.get_characters_character_id_implants(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return implant_type_ids

    def _update_or_create_objs(self, character: Character, implant_type_ids: List[int]):
        if not implant_type_ids:
            self.filter(character=character).delete()
            logger.info("%s: Character has no implants", character)

        EveType.objects.bulk_get_or_create_esi(ids=implant_type_ids)

        # bulk delete & create is fine here, since characters mostly switch
        # between clones, and rarely replace some implants of their current clone
        with transaction.atomic():
            self.filter(character=character).delete()
            implants = [
                self.model(
                    character=character, eve_type=EveType.objects.get(id=eve_type_id)
                )
                for eve_type_id in implant_type_ids
            ]
            logger.info("%s: Storing %s implants", character, len(implants))
            self.bulk_create(implants, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)


class CharacterJumpCloneManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create jump clones for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.JUMP_CLONES,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-clones.read_clones.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token):
        logger.info("%s: Fetching jump clones from ESI", character)
        jump_clones_info = esi.client.Clones.get_characters_character_id_clones(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return jump_clones_info

    # TODO: Replace delete & create with update
    @fetch_token_for_character("esi-universe.read_structures.v1")
    def _update_or_create_objs(
        self, character: Character, token: Token, jump_clones_info: dict
    ):
        from memberaudit.models import CharacterJumpCloneImplant, Location

        self._update_or_create_clone_info(character, token, jump_clones_info)

        jump_clones_list = jump_clones_info.get("jump_clones")
        if jump_clones_list:
            self._prefetch_locations(token, jump_clones_list)

        with transaction.atomic():
            self.filter(character=character).delete()
            if not jump_clones_list:
                logger.info("%s: No jump clones", character)
                return

            logger.info("%s: Storing %s jump clones", character, len(jump_clones_list))
            jump_clones = [
                self.model(
                    character=character,
                    jump_clone_id=record.get("jump_clone_id"),
                    location=get_or_none("location_id", record, Location),
                    name=record.get("name") if record.get("name") else "",
                )
                for record in jump_clones_list
            ]
            self.bulk_create(
                jump_clones, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
            )
            implants = []
            for jump_clone_info in jump_clones_list:
                if jump_clone_info.get("implants"):
                    for implant in jump_clone_info["implants"]:
                        jump_clone = character.jump_clones.get(
                            jump_clone_id=jump_clone_info.get("jump_clone_id")
                        )
                        implants.append(
                            CharacterJumpCloneImplant(
                                jump_clone=jump_clone,
                                eve_type=EveType.objects.get(id=implant),
                            )
                        )

            CharacterJumpCloneImplant.objects.bulk_create(
                implants, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE
            )

    def _update_or_create_clone_info(
        self, character: Character, token: Token, jump_clones_info: dict
    ):
        from memberaudit.models import CharacterCloneInfo, Location

        home_location_data = jump_clones_info.get("home_location")
        if home_location_data and (
            home_location_id := home_location_data.get("location_id")
        ):
            home_location, _ = Location.objects.get_or_create_esi_async(
                id=home_location_id, token=token
            )
        else:
            home_location = None

        return CharacterCloneInfo.objects.update_or_create(
            character=character,
            defaults={
                "home_location": home_location,
                "last_clone_jump_date": jump_clones_info.get("last_clone_jump_date"),
                "last_station_change_date": jump_clones_info.get(
                    "last_station_change_date"
                ),
            },
        )

    def _prefetch_locations(self, token, jump_clones_list):
        """Prefetch related objects ahead of transaction."""

        from memberaudit.models import Location

        incoming_location_ids = {
            record["location_id"]
            for record in jump_clones_list
            if "location_id" in record
        }
        Location.objects.create_missing_esi(incoming_location_ids, token)

        for jump_clone_info in jump_clones_list:
            if jump_clone_info.get("implants"):
                EveType.objects.bulk_get_or_create_esi(
                    ids=jump_clone_info.get("implants", [])
                )


class CharacterLocationManager(models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create location for a character from ESI."""

        return character.update_section_if_changed(
            section=character.UpdateSection.LOCATION,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character(
        ["esi-location.read_location.v1", "esi-universe.read_structures.v1"]
    )
    def _fetch_data_from_esi(self, character: Character, token):
        logger.info("%s: Fetching location from ESI", character)
        location_info = esi.client.Location.get_characters_character_id_location(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()
        return location_info

    @fetch_token_for_character(
        ["esi-location.read_location.v1", "esi-universe.read_structures.v1"]
    )
    def _update_or_create_objs(self, character: Character, token: Token, location_info):
        from memberaudit.models.general import Location

        solar_system_id = location_info["solar_system_id"]
        eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=solar_system_id
        )

        if station_id := location_info.get("station_id"):
            location, _ = Location.objects.get_or_create_esi(id=station_id, token=token)

        elif structure_id := location_info.get("structure_id"):
            location, _ = Location.objects.get_or_create_esi_async(
                id=structure_id, token=token
            )

        else:
            location, _ = Location.objects.get_or_create_esi(
                id=solar_system_id, token=token
            )

        self.update_or_create(
            character=character,
            defaults={"eve_solar_system": eve_solar_system, "location": location},
        )


class CharacterLoyaltyEntryManager(GenericUpdateSimpleObjMixin, models.Manager):
    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create loyalty entries for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.LOYALTY,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-characters.read_loyalty.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> List[dict]:
        logger.info("%s: Fetching loyalty entries from ESI", character)
        loyalty_entries = esi.client.Loyalty.get_characters_character_id_loyalty_points(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        return loyalty_entries

    def _update_or_create_objs(
        self, character: Character, esi_data: List[dict]
    ) -> Set[int]:
        def make_obj_from_esi_entry(character, key, value):
            obj = self.model(
                character=character,
                corporation=EveEntity.objects.get_or_create(id=key)[0],
                loyalty_points=value,
            )
            return obj

        return self._update_or_create_objs_generic(
            character,
            esi_data,
            esi_fields=("corporation_id", "loyalty_points"),
            model_fields=("corporation_id", "loyalty_points"),
            make_obj_from_esi_entry=make_obj_from_esi_entry,
        )


class CharacterMailManager(models.Manager):
    def update_or_create_headers_esi(
        self, character: Character, force_update: bool = False
    ):
        """Update or create mail headers for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.MAILS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
        )

    @fetch_token_for_character("esi-mail.read_mail.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> dict:
        last_mail_id = None
        mail_headers_all = []
        page = 1
        while True:
            logger.info("%s: Fetching mail headers from ESI - page %s", character, page)
            mail_headers = esi.client.Mail.get_characters_character_id_mail(
                character_id=character.eve_character.character_id,
                last_mail_id=last_mail_id,
                token=token.valid_access_token(),
            ).results()

            mail_headers_all += mail_headers
            if len(mail_headers) < 50 or len(mail_headers_all) >= MEMBERAUDIT_MAX_MAILS:
                break

            last_mail_id = min((mail["mail_id"] for mail in mail_headers))
            page += 1

        cutoff_datetime = data_retention_cutoff()
        mail_headers_all_2 = {
            obj["mail_id"]: obj
            for obj in mail_headers_all
            if cutoff_datetime is None
            or not obj.get("timestamp")
            or obj.get("timestamp") > cutoff_datetime
        }
        logger.debug(
            "%s: Received %s mail headers from ESI", character, len(mail_headers_all_2)
        )
        return mail_headers_all_2

    def _update_or_create_objs(self, character: Character, mail_headers):
        if cutoff_datetime := data_retention_cutoff():
            self.filter(character=character, timestamp__lt=cutoff_datetime).delete()

        self._preload_mail_senders(mail_headers=mail_headers)

        with transaction.atomic():
            incoming_ids = set(mail_headers.keys())
            existing_ids = set(
                self.filter(character=character).values_list("mail_id", flat=True)
            )
            create_ids = incoming_ids.difference(existing_ids)
            if create_ids:
                self._create_mail_headers(
                    character=character,
                    mail_headers=mail_headers,
                    create_ids=create_ids,
                )

            update_ids = incoming_ids.difference(create_ids)
            if update_ids:
                self._update_mail_headers(
                    character=character,
                    mail_headers=mail_headers,
                    update_ids=update_ids,
                )

            if not create_ids and not update_ids:
                logger.info("%s: No mails", character)

    def _preload_mail_senders(self, mail_headers):
        from memberaudit.models import MailEntity

        incoming_ids = {o["from"] for o in mail_headers.values()}
        existing_ids = set(MailEntity.objects.values_list("id", flat=True))
        create_ids = incoming_ids.difference(existing_ids)
        for mail_entity_id in create_ids:
            MailEntity.objects.get_or_create_esi_async(id=mail_entity_id)

    def _create_mail_headers(
        self, character: Character, mail_headers: dict, create_ids
    ) -> None:
        from memberaudit.models import MailEntity

        logger.info("%s: Create %s new mail headers", character, len(create_ids))
        new_mail_headers_list = {
            mail_info["mail_id"]: mail_info
            for mail_id, mail_info in mail_headers.items()
            if mail_id in create_ids
        }

        self._add_missing_mailing_lists_from_recipients(
            character=character, new_mail_headers_list=new_mail_headers_list
        )

        # create headers
        new_headers = []
        for mail_id, header in new_mail_headers_list.items():
            new_headers.append(
                self.model(
                    character=character,
                    mail_id=mail_id,
                    sender=get_or_none("from", header, MailEntity),
                    is_read=bool(header.get("is_read")),
                    subject=header.get("subject", ""),
                    timestamp=header.get("timestamp"),
                )
            )

        self.bulk_create(new_headers, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)

        # add recipients and labels
        labels = character.mail_labels.get_all_labels()
        for mail_id, header in new_mail_headers_list.items():
            mail_obj = self.filter(character=character).get(mail_id=mail_id)
            recipients = []
            recipient_type_map = {
                "alliance": MailEntity.Category.ALLIANCE,
                "character": MailEntity.Category.CHARACTER,
                "corporation": MailEntity.Category.CORPORATION,
                "mailing_list": MailEntity.Category.MAILING_LIST,
            }
            for recipient_info in header.get("recipients"):
                recipient, _ = MailEntity.objects.get_or_create(
                    id=recipient_info.get("recipient_id"),
                    defaults={
                        "category": recipient_type_map[
                            recipient_info.get("recipient_type")
                        ]
                    },
                )
                recipients.append(recipient)

            mail_obj.recipients.set(recipients, clear=True)
            MailEntity.objects.bulk_update_names(recipients, keep_names=True)
            self._update_labels_of_mail(
                character=character,
                mail=mail_obj,
                label_ids=header.get("labels"),
                labels=labels,
            )

    def _add_missing_mailing_lists_from_recipients(
        self, character: Character, new_mail_headers_list
    ):
        """Add mailing lists from recipients that are not part of the known
        mailing lists."""
        from memberaudit.models import MailEntity

        incoming_ids = set()
        for header in new_mail_headers_list.values():
            for recipient in header.get("recipients"):
                if recipient.get("recipient_type") == "mailing_list":
                    incoming_ids.add(recipient.get("recipient_id"))

        existing_ids = set(
            MailEntity.objects.filter(
                category=MailEntity.Category.MAILING_LIST
            ).values_list("id", flat=True)
        )
        create_ids = incoming_ids.difference(existing_ids)
        if create_ids:
            logger.info(
                "%s: Adding %s unknown mailing lists from recipients",
                character,
                len(create_ids),
            )
            for list_id in create_ids:
                MailEntity.objects.get_or_create(
                    id=list_id, defaults={"category": MailEntity.Category.MAILING_LIST}
                )

    def _update_labels_of_mail(
        self, character: Character, mail, label_ids: List[int], labels: list
    ) -> None:
        """Update the labels of a mail object from a dict."""
        mail.labels.clear()
        if label_ids:
            labels_to_add = []
            for label_id in label_ids:
                try:
                    labels_to_add.append(labels[label_id])
                except KeyError:
                    logger.info(
                        "%s: Unknown mail label with ID %s for mail %s",
                        character,
                        label_id,
                        mail,
                    )

            mail.labels.add(*labels_to_add)

    def _update_mail_headers(
        self, character: Character, mail_headers: dict, update_ids
    ) -> None:
        logger.info("%s: Updating %s mail headers", character, len(update_ids))
        mail_pks = self.filter(character=character, mail_id__in=update_ids).values_list(
            "pk", flat=True
        )
        labels = character.mail_labels.get_all_labels()
        mails = self.in_bulk(mail_pks)
        for mail in mails.values():
            mail_header = mail_headers.get(mail.mail_id)
            if mail_header:
                mail.is_read = bool(mail_header.get("is_read"))
                self._update_labels_of_mail(
                    character=character,
                    mail=mail,
                    label_ids=mail_header.get("labels"),
                    labels=labels,
                )

        self.bulk_update(mails.values(), ["is_read"])

    def update_or_create_body_esi(
        self, character: Character, mail: CharacterMail, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create mail body for a character from ESI."""
        try:
            mail_body = self._fetch_mail_body_from_esi(character, mail)
        except HTTPNotFound:
            logger.info(
                "%s: Mail %s was deleted in game. Removing mail header.",
                character,
                mail,
            )
            mail.delete()
            return UpdateSectionResult(is_changed=True, is_updated=True)

        is_changed = mail.body != mail_body
        if is_changed or force_update:
            mail.body = mail_body
            mail.save()
            is_updated = True
            eve_xml_to_html(mail.body)  # resolve names early
        else:
            is_updated = False

        store_character_data_to_disk_when_enabled(
            character=character,
            data=mail_body,
            section="mails",
            suffix=f"body-{mail.mail_id}",
        )

        return UpdateSectionResult(is_changed=is_changed, is_updated=is_updated)

    def _fetch_mail_body_from_esi(
        self, character: Character, mail: CharacterMail
    ) -> str:
        logger.info(
            "%s: Fetching mail body from ESI for mail ID %d", character, mail.mail_id
        )
        token = character.fetch_token("esi-mail.read_mail.v1")
        mail_body = esi.client.Mail.get_characters_character_id_mail_mail_id(
            character_id=character.eve_character.character_id,
            mail_id=mail.mail_id,
            token=token.valid_access_token(),
        ).result()

        return mail_body.get("body", "")


class CharacterMailLabelManager(models.Manager):
    def get_all_labels(self) -> Dict[int, Any]:
        """Return all label objects as dict by label_id."""
        label_pks = self.values_list("pk", flat=True)
        return {label.label_id: label for label in self.in_bulk(label_pks).values()}

    def update_or_create_esi(
        self, character: Character, force_update: bool = False
    ) -> UpdateSectionResult:
        """Update or create mail labels for a character from ESI."""
        return character.update_section_if_changed(
            section=character.UpdateSection.MAILS,
            fetch_func=self._fetch_data_from_esi,
            store_func=self._update_or_create_objs,
            force_update=force_update,
            hash_num=3,
        )

    @fetch_token_for_character("esi-mail.read_mail.v1")
    def _fetch_data_from_esi(self, character: Character, token: Token) -> dict:
        from memberaudit.models import CharacterMailUnreadCount

        logger.info("%s: Fetching mail labels from ESI", character)
        mail_labels_info = esi.client.Mail.get_characters_character_id_mail_labels(
            character_id=character.eve_character.character_id,
            token=token.valid_access_token(),
        ).results()

        if mail_labels_info.get("total_unread_count"):
            CharacterMailUnreadCount.objects.update_or_create(
                character=character,
                defaults={"total": mail_labels_info.get("total_unread_count")},
            )

        mail_labels = mail_labels_info.get("labels")
        if not mail_labels:
            return {}
        return {obj["label_id"]: obj for obj in mail_labels if "label_id" in obj}

    @transaction.atomic()
    def _update_or_create_objs(self, character: Character, mail_labels_list: dict):
        logger.info("%s: Storing %s mail labels", character, len(mail_labels_list))
        incoming_ids = set(mail_labels_list.keys())
        existing_ids = set(
            self.filter(character=character).values_list("label_id", flat=True)
        )
        obsolete_ids = existing_ids.difference(incoming_ids)
        if obsolete_ids:
            self.filter(character=character, label_id__in=obsolete_ids).delete()

        create_ids = incoming_ids.difference(existing_ids)
        if create_ids:
            self._create_new_mail_labels(
                character=character,
                mail_labels_list=mail_labels_list,
                label_ids=create_ids,
            )

        update_ids = incoming_ids.difference(create_ids)
        if update_ids:
            self._update_existing_mail_labels(
                character=character,
                mail_labels_list=mail_labels_list,
                label_ids=update_ids,
            )

    def _create_new_mail_labels(
        self, character: Character, mail_labels_list: dict, label_ids: set
    ):
        new_labels = [
            self.model(
                character=character,
                label_id=label.get("label_id"),
                color=label.get("color"),
                name=label.get("name"),
                unread_count=label.get("unread_count"),
            )
            for label_id, label in mail_labels_list.items()
            if label_id in label_ids
        ]
        self.bulk_create(new_labels, batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE)

    def _update_existing_mail_labels(
        self, character: Character, mail_labels_list: dict, label_ids: set
    ):
        logger.info("%s: Updating %s mail labels", character, len(label_ids))
        update_pks = list(
            self.filter(character=character, label_id__in=label_ids).values_list(
                "pk", flat=True
            )
        )
        labels = self.in_bulk(update_pks)
        for label in labels.values():
            record = mail_labels_list.get(label.label_id)
            if record:
                label.name = record.get("name")
                label.color = record.get("color")
                label.unread_count = record.get("unread_count")

        self.bulk_update(
            labels.values(),
            fields=["name", "color", "unread_count"],
            batch_size=MEMBERAUDIT_BULK_METHODS_BATCH_SIZE,
        )



"""Logic for standings in Eve Online."""

from django.db import models
from django.utils.translation import gettext_lazy as _


class Standing(models.IntegerChoices):
    """An Eve Online standing."""

    EXCELLENT = 10, _("excellent standing")
    GOOD = 5, _("good standing")
    NEUTRAL = 0, _("neutral standing")
    BAD = -5, _("bad standing")
    TERRIBLE = -10, _("terrible standing")

    @classmethod
    def from_value(cls, value: float) -> "Standing":
        """Create new objects from standing number."""
        if value > 5:
            return cls.EXCELLENT

        if 5 >= value > 0:
            return cls.GOOD

        if value == 0:
            return cls.NEUTRAL

        if 0 > value >= -5:
            return cls.BAD

        return cls.TERRIBLE


def calc_effective_standing(
    unadjusted_standing, skill_level, skill_modifier, max_possible_standing
):
    """Calculate effective after applying skill."""
    effective_standing = unadjusted_standing + (
        (max_possible_standing - unadjusted_standing) * skill_modifier * skill_level
    )
    return effective_standing



"""Helpers for models."""

from __future__ import annotations

import datetime as dt
import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.timezone import now

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from memberaudit import __title__
from memberaudit.app_settings import (
    MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS,
    MEMBERAUDIT_STORE_ESI_DATA_ENABLED,
    MEMBERAUDIT_STORE_ESI_DATA_SECTIONS,
)

if TYPE_CHECKING:
    from memberaudit.models import Character

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def store_character_data_to_disk_when_enabled(
    character: Character, data: Any, section: Character.UpdateSection, suffix: str = ""
) -> Optional[Path]:
    """Store character data to disk in JSON files for debugging when enabled.

    Returns path to created data file.
    """
    from memberaudit.models import Character

    if not MEMBERAUDIT_STORE_ESI_DATA_ENABLED:
        return None

    try:
        section_obj = Character.UpdateSection(section)
    except ValueError:
        logger.exception("%s: Failed to write debug data: %s", character, section)
        return None

    if (
        MEMBERAUDIT_STORE_ESI_DATA_SECTIONS
        and section_obj.value not in MEMBERAUDIT_STORE_ESI_DATA_SECTIONS
    ):
        return None

    if (
        MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS
        and character.id not in MEMBERAUDIT_STORE_ESI_DATA_CHARACTERS
    ):
        return None

    path = _create_path_if_it_not_exists()
    file_path = _generate_file_path(character, section_obj, suffix, path)
    success = _write_data(data, file_path)
    if not success:
        return None

    return file_path


def _create_path_if_it_not_exists() -> Path:
    today_str = now().strftime("%Y%m%d")
    path = Path(settings.BASE_DIR) / "temp" / "memberaudit_log" / today_str
    path.mkdir(parents=True, exist_ok=True)
    return path


def _generate_file_path(
    character: Character, section: Character.UpdateSection, suffix: str, path: Path
) -> Path:
    now_str = now().strftime("%Y%m%d%H%M")
    name = f"{section.value}_{suffix}" if suffix else section.value
    file_name = f"character_{character.pk}_{name}_{now_str}.json"
    file_path = path / file_name
    return file_path


def _write_data(data, file_path: Path) -> bool:
    try:
        with file_path.open("w", encoding="utf-8") as file:
            json.dump(data, file, cls=DjangoJSONEncoder, sort_keys=True, indent=4)

    except (OSError, TypeError):
        logger.exception("Failed to write debug data to: %s", file_path)
        return False

    logger.info("Wrote debug data to: %s", file_path)
    return True


class AddGenericReprMixin:
    """Mixin adds a generic repr to a Django model."""

    def __repr__(self) -> str:
        fields = {}
        for field in self._meta.get_fields():
            if field.one_to_many or field.many_to_many:
                continue

            try:
                value = getattr(self, field.attname)
            except AttributeError:
                continue

            if isinstance(value, dt.datetime):
                value = f"<{value}>"

            fields[field.attname] = value

        fields_2 = dict(sorted(fields.items()))
        fields_str = ", ".join([f"{key}={value}" for key, value in fields_2.items()])
        class_name = type(self).__name__
        return f"{class_name}({fields_str})"



# flake8: noqa
"""
This is a standalone scripts that generates a test character
"""

import inspect
import os
import sys

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
myauth_dir = (
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(currentdir))))
    + "/myauth"
)
sys.path.insert(0, myauth_dir)


import django
from django.apps import apps

# init and setup django project
print("Initializing Django...")
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

if not apps.is_installed("memberaudit"):
    raise RuntimeError("The app memberaudit is not installed")

import datetime as dt
import random

from django.core.exceptions import ObjectDoesNotExist
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveSolarSystem

from allianceauth.eveonline.models import EveCharacter

from memberaudit.models import (
    Character,
    CharacterDetails,
    CharacterLocation,
    CharacterWalletJournalEntry,
)
from memberaudit.tests.testdata.factories import (
    create_character_planet,
    create_character_wallet_journal_entry,
)
from memberaudit.tests.utils import create_memberaudit_character

WALLET_JOURNAL_ENTRIES = 1_000


def main():
    eve_character, corporation, character = recreate_character()
    create_details(eve_character, corporation, character)
    create_location(character)
    create_wallet_journal(eve_character, character)
    create_planets(character)
    print("DONE")


def recreate_character():
    print("Recreating character...")
    eve_character, _ = EveCharacter.objects.get_or_create(
        character_id=92532650,
        defaults={
            "character_name": "CCP Falcon",
            "corporation_id": 1000127,
            "corporation_name": "Guristas",
            "corporation_ticker": "XXX",
        },
    )
    corporation, _ = EveEntity.objects.get_or_create_esi(
        id=eve_character.corporation_id
    )
    try:
        eve_character.character_ownership.user.delete()
    except ObjectDoesNotExist:
        pass
    Character.objects.get(eve_character=eve_character).delete()
    character = create_memberaudit_character(92532650)
    return eve_character, corporation, character


def create_details(eve_character, corporation, character):
    CharacterDetails.objects.create(
        character=character,
        birthday=now(),
        corporation=corporation,
        description="This is a generated test character.",
        eve_bloodline_id=1,
        eve_race_id=1,
        name=eve_character.character_name,
    )


def create_location(character):
    eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
        id=30004984, include_children=True
    )
    CharacterLocation.objects.create(
        character=character, eve_solar_system=eve_solar_system
    )


def create_wallet_journal(eve_character, character):
    print(f"Creating wallet journal with {WALLET_JOURNAL_ENTRIES} entries...")
    first_party, _ = EveEntity.objects.get_or_create_esi(id=eve_character.character_id)
    some_characters = list(
        EveEntity.objects.filter(category=EveEntity.CATEGORY_CHARACTER).order_by("?")[
            :20
        ]
    )
    objs = []
    balance = 10_000_000_000_000
    date = now() - dt.timedelta(days=5_000)
    for entry_id in range(1, WALLET_JOURNAL_ENTRIES):
        amount = random.randint(1_000_000, 100_000_000)
        balance -= amount
        second_party = random.choice(some_characters)
        date += dt.timedelta(minutes=random.randint(1, 15))
        objs.append(
            CharacterWalletJournalEntry(
                character=character,
                entry_id=entry_id,
                amount=amount,
                balance=balance,
                context_id_type=CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
                date=date,
                description=f"Generated entry #{entry_id}",
                first_party=first_party,
                second_party=second_party,
                ref_type="player_donation",
            )
        )
    CharacterWalletJournalEntry.objects.bulk_create(objs, batch_size=500)


def create_planets(character):
    for _ in range(6):
        create_character_planet(character)


main()



"""Manager for Character model."""
# pylint: disable=missing-class-docstring

from typing import Set

from django.contrib.auth.models import Permission, User
from django.db import models
from django.db.models import Case, Count, Q, Value, When

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger
from app_utils.caching import ObjectCacheMixin
from app_utils.django import users_with_permission
from app_utils.logging import LoggerAddTag

from memberaudit import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class CharacterQuerySet(models.QuerySet):
    def eve_character_ids(self) -> Set[int]:
        """Return EveCharacter IDs of all characters in this QuerySet."""
        return set(self.values_list("eve_character__character_id", flat=True))

    def owned_by_user(self, user: User) -> models.QuerySet:
        """Filter character owned by user."""
        return self.filter(eve_character__character_ownership__user__pk=user.pk)

    def annotate_total_update_status(self) -> models.QuerySet:
        """Add total_update_status annotations."""
        from memberaudit.models import Character

        enabled_sections = list(Character.UpdateSection.enabled_sections())
        num_sections_total = len(enabled_sections)
        qs = (
            self.annotate(
                num_sections_total=Count(
                    "update_status_set",
                    filter=Q(update_status_set__section__in=enabled_sections),
                )
            )
            .annotate(
                num_sections_ok=Count(
                    "update_status_set",
                    filter=Q(
                        update_status_set__section__in=enabled_sections,
                        update_status_set__is_success=True,
                    ),
                )
            )
            .annotate(
                num_sections_failed=Count(
                    "update_status_set",
                    filter=Q(
                        update_status_set__section__in=enabled_sections,
                        update_status_set__is_success=False,
                    ),
                )
            )
            .annotate(
                num_sections_token_error=Count(
                    "update_status_set",
                    filter=Q(
                        update_status_set__section__in=enabled_sections,
                        update_status_set__has_token_error=True,
                    ),
                )
            )
            .annotate(
                total_update_status=Case(
                    When(
                        is_disabled=True,
                        then=Value(Character.TotalUpdateStatus.DISABLED.value),
                    ),
                    When(
                        num_sections_token_error=1,
                        then=Value(Character.TotalUpdateStatus.LIMITED_TOKEN.value),
                    ),
                    When(
                        num_sections_failed__gt=0,
                        then=Value(Character.TotalUpdateStatus.ERROR.value),
                    ),
                    When(
                        num_sections_ok=num_sections_total,
                        then=Value(Character.TotalUpdateStatus.OK.value),
                    ),
                    When(
                        num_sections_total__lt=num_sections_total,
                        then=Value(Character.TotalUpdateStatus.INCOMPLETE.value),
                    ),
                    default=Value(Character.TotalUpdateStatus.IN_PROGRESS.value),
                )
            )
        )
        return qs

    def disable_characters_with_no_owner(self) -> int:
        """Disable characters which have no owner. Return count of disabled characters."""
        orphaned_characters = self.filter(
            eve_character__character_ownership__isnull=True, is_disabled=False
        )
        if orphaned_characters.exists():
            orphans = list(
                orphaned_characters.values_list(
                    "eve_character__character_name", flat=True
                ).order_by("eve_character__character_name")
            )
            orphaned_characters.update(is_disabled=True)
            logger.info(
                "Disabled %d characters which do not belong to a user: %s",
                len(orphans),
                ", ".join(orphans),
            )
            return len(orphans)

        return 0


class CharacterManagerBase(ObjectCacheMixin, models.Manager):
    def characters_of_user_to_register_count(self, user: User) -> int:
        """Return count of a users's characters known to Auth,
        which needs to be (re-)registered.
        """
        unregistered = CharacterOwnership.objects.filter(
            user=user, character__memberaudit_character__isnull=True
        ).count()
        enabled_sections = list(self.model.UpdateSection.enabled_sections())
        token_errors = (
            self.filter(eve_character__character_ownership__user=user)
            .filter(
                Q(
                    update_status_set__section__in=enabled_sections,
                    update_status_set__has_token_error=True,
                )
                | Q(is_disabled=True),
            )
            .distinct()
            .count()
        )
        return unregistered + token_errors

    def user_has_scope(self, user: User) -> models.QuerySet:
        """Return characters the given user has the scope permission to access."""
        if user.has_perm("memberaudit.view_everything"):
            return self.all()
        qs = self.filter(eve_character__character_ownership__user=user)
        if (
            user.has_perm("memberaudit.view_same_alliance")
            and user.profile.main_character.alliance_id
        ):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__alliance_id=(
                    user.profile.main_character.alliance_id
                )
            )
        elif user.has_perm("memberaudit.view_same_corporation"):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__corporation_id=(
                    user.profile.main_character.corporation_id
                )
            )
        return qs

    def user_has_access(self, user: User) -> models.QuerySet:
        """Return characters the given user has permission to access
        via character viewer.
        """
        if user.has_perm("memberaudit.view_everything") and user.has_perm(
            "memberaudit.characters_access"
        ):
            return self.all()
        qs = self.filter(eve_character__character_ownership__user=user)
        if (
            user.has_perm("memberaudit.characters_access")
            and user.has_perm("memberaudit.view_same_alliance")
            and user.profile.main_character.alliance_id
        ):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__alliance_id=(
                    user.profile.main_character.alliance_id
                )
            )
        elif user.has_perm("memberaudit.characters_access") and user.has_perm(
            "memberaudit.view_same_corporation"
        ):
            qs |= self.filter(
                eve_character__character_ownership__user__profile__main_character__corporation_id=(
                    user.profile.main_character.corporation_id
                )
            )
        if user.has_perm("memberaudit.view_shared_characters"):
            permission_to_share_characters = Permission.objects.select_related(
                "content_type"
            ).get(
                content_type__app_label=self.model._meta.app_label,
                codename="share_characters",
            )
            viewable_users = users_with_permission(permission_to_share_characters)
            qs |= self.filter(
                is_shared=True,
                eve_character__character_ownership__user__in=viewable_users,
            )
        return qs


CharacterManager = CharacterManagerBase.from_queryset(CharacterQuerySet)


class CharacterUpdateStatusQuerySet(models.QuerySet):
    def filter_enabled_sections(self) -> models.QuerySet:
        """Filter enabled sections."""
        from memberaudit.models import Character

        enabled_sections = list(Character.UpdateSection.enabled_sections())
        return self.filter(section__in=enabled_sections)


class CharacterUpdateStatusManagerBase(models.Manager):
    pass


CharacterUpdateStatusManager = CharacterUpdateStatusManagerBase.from_queryset(
    CharacterUpdateStatusQuerySet
)



"""Old style asset tests."""

import datetime as dt
from typing import Dict
from unittest.mock import patch

from bravado.exception import HTTPError
from celery.exceptions import Retry as CeleryRetry

from django.test import TestCase, override_settings, tag
from django.utils.timezone import now
from esi.models import Token
from eveuniverse.models import EveSolarSystem, EveType
from eveuniverse.tests.testdata.factories import create_eve_entity

from allianceauth.eveonline.models import EveCharacter
from app_utils.esi import EsiErrorLimitExceeded, EsiOffline, EsiStatus
from app_utils.esi_testing import build_http_error
from app_utils.testing import (  # NoSocketsTestCase,
    create_authgroup,
    create_user_from_evecharacter,
    generate_invalid_pk,
)

from memberaudit import tasks
from memberaudit.helpers import UpdateSectionResult
from memberaudit.models import Character, CharacterUpdateStatus, Location

from .testdata.esi_client_stub import esi_client_stub, esi_error_stub, esi_stub
from .testdata.factories import (
    create_character,
    create_character_asset,
    create_character_update_status,
    create_compliance_group_designation,
)
from .testdata.load_entities import load_entities
from .testdata.load_eveuniverse import load_eveuniverse
from .testdata.load_locations import load_locations
from .utils import create_memberaudit_character, reset_celery_once_locks

MODELS_PATH = "memberaudit.models"
MANAGERS_PATH = "memberaudit.managers"
TASKS_PATH = "memberaudit.tasks"


@patch(TASKS_PATH + ".update_compliance_groups_for_all", spec=True)
@patch(TASKS_PATH + ".update_all_characters", spec=True)
@patch(TASKS_PATH + ".update_market_prices", spec=True)
class TestRegularUpdates(TestCase):
    def test_should_run_update_for_all_except_compliance_groups(
        self,
        mock_update_market_prices,
        mock_update_all_characters,
        mock_update_compliance_groups_for_all,
    ):
        # when
        tasks.run_regular_updates()
        # then
        self.assertTrue(mock_update_market_prices.apply_async.called)
        self.assertTrue(mock_update_all_characters.apply_async.called)
        self.assertFalse(mock_update_compliance_groups_for_all.apply_async.called)

    def test_should_run_update_for_all_incl_compliance_groups(
        self,
        mock_update_market_prices,
        mock_update_all_characters,
        mock_update_compliance_groups_for_all,
    ):
        # given
        group = create_authgroup(internal=False)
        create_compliance_group_designation(group)
        # when
        tasks.run_regular_updates()
        # then
        self.assertTrue(mock_update_market_prices.apply_async.called)
        self.assertTrue(mock_update_all_characters.apply_async.called)
        self.assertTrue(mock_update_compliance_groups_for_all.apply_async.called)


@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_1.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_2.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_3.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_1.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_2.esi", esi_stub)
@patch(MANAGERS_PATH + ".character_sections_3.esi", esi_stub)
@patch(MANAGERS_PATH + ".general.esi", esi_stub)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestUpdateCharacter(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        reset_celery_once_locks()

    def setUp(self) -> None:
        self.character_1001 = create_memberaudit_character(1001)

    @tag("breaks_with_tox")  # FIXME: Find solution
    @patch(MODELS_PATH + ".characters.MEMBERAUDIT_FEATURE_ROLES_ENABLED", True)
    def test_should_update_all_sections_from_scratch(self):
        # when
        result = tasks.update_character(self.character_1001.pk)

        # then
        self.assertTrue(result)

        status_all: Dict[str, CharacterUpdateStatus] = {
            obj.section: obj for obj in self.character_1001.update_status_set.all()
        }
        for section in Character.UpdateSection:
            with self.subTest(section=section):
                self.assertTrue(status_all[section].is_success)

    @tag("breaks_with_tox")  # FXME: Find solution
    @patch(MODELS_PATH + ".characters.MEMBERAUDIT_FEATURE_ROLES_ENABLED", False)
    def test_should_update_enabled_sections_only(self):
        # given
        run_started_at = now() - dt.timedelta(hours=24)
        for section in Character.UpdateSection:
            create_character_update_status(
                character=self.character_1001,
                section=section,
                is_success=True,
                run_started_at=run_started_at,
                run_finished_at=run_started_at,
            )

        # when
        result = tasks.update_character(self.character_1001.pk)

        # then
        self.assertTrue(result)

        for section in Character.UpdateSection.enabled_sections():
            with self.subTest(section=section):
                self.assertFalse(
                    self.character_1001.is_update_needed_for_section(section=section)
                )

        self.assertTrue(
            self.character_1001.is_update_needed_for_section(
                section=Character.UpdateSection.ROLES
            )
        )

    @patch(TASKS_PATH + ".Character.update_loyalty", spec=True)
    def test_should_update_normal_section_only_when_stale(self, update_loyalty):
        # given
        create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.LOYALTY,
            is_success=True,
            run_started_at=now() - dt.timedelta(seconds=30),
            run_finished_at=now(),
        )
        # when
        tasks.update_character(self.character_1001.pk)
        # then
        self.assertFalse(update_loyalty.called)

    @patch(TASKS_PATH + ".update_character_mails", spec=True)
    def test_should_update_special_section_only_when_stale(
        self, mock_update_character_mails
    ):
        # given
        create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.MAILS,
            is_success=True,
            run_started_at=now() - dt.timedelta(seconds=30),
            run_finished_at=now(),
        )
        # when
        tasks.update_character(self.character_1001.pk)
        # then
        self.assertFalse(mock_update_character_mails.apply_async.called)

    def test_should_update_section_when_not_stale_but_force_update_requested(self):
        # given
        status = create_character_update_status(
            character=self.character_1001,
            section=Character.UpdateSection.SKILLS.value,
            is_success=True,
            run_started_at=now() - dt.timedelta(seconds=30),
            run_finished_at=now(),
        )
        last_finished = status.run_finished_at
        # when
        tasks.update_character(
            self.character_1001.pk, force_update=True, ignore_stale=True
        )
        # then
        status.refresh_from_db()
        self.assertGreater(status.run_finished_at, last_finished)
        self.assertTrue(status.update_finished_at)

    def test_should_update_section_when_not_stale_but_ignore_stale(self):
        # given
        tasks.update_character_skills(self.character_1001.pk, force_update=True)
        status = self.character_1001.update_status_for_section("skills")
        status.update_finished_at = None
        status.save()
        # when
        tasks.update_character(self.character_1001.pk, ignore_stale=True)
        # then
        status.refresh_from_db()
        self.assertIsNone(status.update_finished_at)

    def test_should_do_nothing_when_not_required(self):
        # given
        for section in Character.UpdateSection.values:
            create_character_update_status(
                character=self.character_1001,
                section=section,
                is_success=True,
                run_started_at=now() - dt.timedelta(seconds=30),
                run_finished_at=now(),
            )
        # when
        result = tasks.update_character(self.character_1001.pk)
        # then
        self.assertFalse(result)

    def test_can_do_forced_update(self):
        # when
        result = tasks.update_character(self.character_1001.pk, force_update=True)
        # then
        self.assertTrue(result)
        self.assertTrue(self.character_1001.is_update_status_ok())

    def test_skip_update_for_orphans(self):
        # given
        character = create_character(EveCharacter.objects.get(character_id=1121))
        # when
        result = tasks.update_character(character.pk)
        # then
        self.assertFalse(result)
        self.assertIsNone(character.is_update_status_ok())


@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_1.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_2.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_3.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".character_sections_1.esi", esi_error_stub)
@patch(MANAGERS_PATH + ".character_sections_2.esi", esi_error_stub)
@patch(MANAGERS_PATH + ".character_sections_3.esi", esi_error_stub)
@patch(MANAGERS_PATH + ".general.esi", esi_error_stub)
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestUpdateCharacterErrorReporting(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        reset_celery_once_locks()

    def setUp(self) -> None:
        self.character_1001 = create_memberaudit_character(1001)

    def test_should_report_errors_during_updates(self):
        # when
        with self.assertRaises(HTTPError):  # raised when trying to fetch attributes
            tasks.update_character(self.character_1001.pk)
        # then
        self.assertFalse(self.character_1001.is_update_status_ok())
        status = self.character_1001.update_status_set.filter(
            character=self.character_1001, is_success=False
        ).first()
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")
        self.assertTrue(status.run_finished_at)


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(MANAGERS_PATH + ".character_sections_1.esi")
class TestUpdateCharacterAssets(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.character_1002 = create_memberaudit_character(1002)
        cls.token = (
            cls.character_1001.eve_character.character_ownership.user.token_set.first()
        )
        cls.jita = EveSolarSystem.objects.get(id=30000142)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.amamake = EveSolarSystem.objects.get(id=30002537)
        cls.structure_1 = Location.objects.get(id=1_000_000_000_001)
        reset_celery_once_locks()

    def test_should_create_assets_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        tasks.update_character_assets(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
                1_100_000_000_007,
                1_100_000_000_008,
            },
        )

        asset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        self.assertTrue(asset.is_blueprint_copy)
        self.assertTrue(asset.is_singleton)
        self.assertEqual(asset.location_flag, "Hangar")
        self.assertEqual(asset.location_id, 60003760)
        self.assertEqual(asset.quantity, 1)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=20185))
        self.assertEqual(asset.name, "Parent Item 1")

        asset = self.character_1001.assets.get(item_id=1_100_000_000_002)
        self.assertFalse(asset.is_blueprint_copy)
        self.assertTrue(asset.is_singleton)
        self.assertEqual(asset.location_flag, "???")
        self.assertEqual(asset.parent.item_id, 1_100_000_000_001)
        self.assertEqual(asset.quantity, 1)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=19540))
        self.assertEqual(asset.name, "Leaf Item 2")

        asset = self.character_1001.assets.get(item_id=1_100_000_000_003)
        self.assertEqual(asset.parent.item_id, 1_100_000_000_001)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=23))

        asset = self.character_1001.assets.get(item_id=1_100_000_000_004)
        self.assertEqual(asset.parent.item_id, 1_100_000_000_003)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=19553))

        asset = self.character_1001.assets.get(item_id=1_100_000_000_005)
        self.assertEqual(asset.location, self.structure_1)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=20185))

        asset = self.character_1001.assets.get(item_id=1_100_000_000_006)
        self.assertEqual(asset.parent.item_id, 1_100_000_000_005)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=19540))

        asset = self.character_1001.assets.get(item_id=1_100_000_000_007)
        self.assertEqual(asset.location_id, 30000142)
        self.assertEqual(asset.name, "")
        self.assertEqual(asset.eve_type, EveType.objects.get(id=19540))

        asset = self.character_1001.assets.get(item_id=1_100_000_000_008)
        self.assertEqual(asset.location_id, 1_000_000_000_001)

    def test_should_remove_obsolete_assets(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_asset(
            character=self.character_1001, item_id=1100000000666, location=self.jita_44
        )

        # when
        tasks.update_character_assets(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
                1_100_000_000_007,
                1_100_000_000_008,
            },
        )

    def test_should_update_existing_assets(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_asset(
            character=self.character_1001,
            item_id=1_100_000_000_001,
            location=self.jita_44,
            eve_type=EveType.objects.get(id=20185),
            is_singleton=True,
            name="Parent Item 1",
            quantity=10,
        )

        # when
        tasks.update_character_assets(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
                1_100_000_000_007,
                1_100_000_000_008,
            },
        )

        asset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        self.assertTrue(asset.is_singleton)
        self.assertEqual(asset.location_id, 60003760)
        self.assertEqual(asset.quantity, 1)
        self.assertEqual(asset.eve_type, EveType.objects.get(id=20185))
        self.assertEqual(asset.name, "Parent Item 1")

    def test_should_keep_assets_which_are_moved_to_different_locations(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        parent_asset = create_character_asset(
            character=self.character_1001,
            item_id=1100000000666,
            location=self.jita_44,
            eve_type=EveType.objects.get(id=20185),
        )
        create_character_asset(
            character=self.character_1001,
            item_id=1_100_000_000_002,
            parent=parent_asset,
            eve_type=EveType.objects.get(id=19540),
            quantity=1,
        )

        # when
        tasks.update_character_assets(self.character_1001.pk, True)

        # then
        self.assertSetEqual(
            self.character_1001.assets.item_ids(),
            {
                1_100_000_000_001,
                1_100_000_000_002,
                1_100_000_000_003,
                1_100_000_000_004,
                1_100_000_000_005,
                1_100_000_000_006,
                1_100_000_000_007,
                1_100_000_000_008,
            },
        )

    def test_should_report_update_success_when_update_succeeded(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        tasks.update_character_assets(self.character_1001.pk, True)

        # then
        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.ASSETS
        )
        self.assertTrue(status.is_success)
        self.assertFalse(status.error_message)

    def test_should_report_the_error_when_update_failed(self, mock_esi):
        # given
        exception = build_http_error(502, "Test exception")
        mock_esi.client.Assets.get_characters_character_id_assets.side_effect = (
            exception
        )

        # when
        with self.assertRaises(HTTPError):
            tasks.update_character_assets(self.character_1001.pk, True)

        # then
        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.ASSETS
        )
        self.assertFalse(status.is_success)
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")

    def test_should_report_error_when_preload_objects_failed(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        with patch(
            MANAGERS_PATH + ".general.LocationManager.get_or_create_esi_async",
            spec=True,
        ) as m:
            exception = build_http_error(502, "Test exception")
            m.side_effect = exception
            with self.assertRaises(HTTPError):
                tasks.update_character_assets(self.character_1001.pk, True)

        # then
        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.ASSETS
        )
        self.assertFalse(status.is_success)
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")

    def test_should_report_the_error_when_building_the_asset_tree_failed(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub

        # when
        with patch(MANAGERS_PATH + ".character_sections_1.logger") as m:
            exception = build_http_error(502, "Test exception")
            m.info.side_effect = exception
            with self.assertRaises(HTTPError):
                tasks.update_character_assets(self.character_1001.pk, True)

        # then
        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.ASSETS
        )
        self.assertFalse(status.is_success)
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")

    def test_should_not_recreate_asset_tree_when_info_from_ESI_is_unchanged(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        self.character_1001.reset_update_section(Character.UpdateSection.ASSETS)
        tasks.update_character_assets(self.character_1001.pk, True)
        asset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        asset.name = "New Name"
        asset.save()

        # when
        tasks.update_character_assets(self.character_1001.pk, False)

        # then
        asset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        self.assertEqual(asset.name, "New Name")

        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.ASSETS
        )
        self.assertTrue(status.is_success)

    def test_should_recreate_asset_tree_when_info_from_ESI_is_unchanged_and_is_forced(
        self, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        self.character_1001.reset_update_section(Character.UpdateSection.ASSETS)
        tasks.update_character_assets(self.character_1001.pk, True)
        asset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        asset.name = "New Name"
        asset.save()

        # when
        tasks.update_character_assets(self.character_1001.pk, force_update=True)

        # then
        asset = self.character_1001.assets.get(item_id=1_100_000_000_001)
        self.assertEqual(asset.name, "Parent Item 1")

        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.ASSETS
        )
        self.assertTrue(status.is_success)


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_1.esi")
class TestUpdateCharacterContacts(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.token = (
            cls.character_1001.eve_character.character_ownership.user.token_set.first()
        )
        reset_celery_once_locks()

    def test_update_ok(self, mock_esi):
        """when update succeeded then report update success"""
        mock_esi.client = esi_client_stub

        tasks.update_character_contacts(self.character_1001.pk, True)

        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.CONTACTS
        )
        self.assertTrue(status.is_success)
        self.assertFalse(status.error_message)

    def test_detect_error(self, mock_esi):
        """when update failed then report the error"""
        exception = build_http_error(502, "Test exception")
        mock_esi.client.Contacts.get_characters_character_id_contacts_labels.side_effect = (
            exception
        )

        with self.assertRaises(HTTPError):
            tasks.update_character_contacts(self.character_1001.pk, True)

        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.CONTACTS
        )
        self.assertFalse(status.is_success)
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_1.esi")
class TestUpdateCharacterContracts(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.token = (
            cls.character_1001.eve_character.character_ownership.user.token_set.first()
        )
        reset_celery_once_locks()

    def test_should_record_success_when_update_completed_successfully(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        tasks.update_character_contracts(self.character_1001.pk, True)

        # then
        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.CONTRACTS
        )
        self.assertTrue(status.is_success)
        self.assertFalse(status.error_message)
        self.assertTrue(status.run_finished_at)
        self.assertTrue(status.update_finished_at)

    def test_should_record_error_when_update_failed(self, mock_esi):
        # given
        exception = build_http_error(502, "Test exception")
        mock_esi.client.Contracts.get_characters_character_id_contracts.side_effect = (
            exception
        )

        # when
        with self.assertRaises(HTTPError):
            tasks.update_character_contracts(self.character_1001.pk, True)

        # then
        status = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.CONTRACTS
        )
        self.assertFalse(status.is_success)
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
@patch(MANAGERS_PATH + ".character_sections_2.data_retention_cutoff", lambda: None)
@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(MANAGERS_PATH + ".character_sections_2.esi")
@patch(MANAGERS_PATH + ".general.esi")
class TestUpdateCharacterMails(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character_1001 = create_memberaudit_character(1001)
        cls.token = (
            cls.character_1001.eve_character.character_ownership.user.token_set.first()
        )
        reset_celery_once_locks()

    def test_should_report_success_when_update_is_completed_successfully(
        self, mock_esi_character, mock_esi_sections
    ):
        # given
        mock_esi_character.client = esi_client_stub
        mock_esi_sections.client = esi_client_stub

        # when
        tasks.update_character_mails.delay(self.character_1001.pk, True)

        # then
        status: CharacterUpdateStatus = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.MAILS
        )
        self.assertTrue(status.is_success)
        self.assertFalse(status.error_message)
        self.assertTrue(status.run_started_at)
        self.assertTrue(status.run_finished_at)
        self.assertTrue(status.update_started_at)
        self.assertTrue(status.update_finished_at)
        mail_ids = set(self.character_1001.mails.values_list("mail_id", flat=True))
        self.assertSetEqual(mail_ids, {1, 2, 3})
        mail = self.character_1001.mails.get(mail_id=1)
        self.assertEqual(mail.subject, "Mail 1")
        self.assertEqual(mail.body, "blah blah blah 😓")

    # TODO: Add test to check force update works

    def test_should_report_error_when_update_failed(
        self, mock_esi_character, mock_esi_sections
    ):
        # given
        exception = build_http_error(502, "Test exception")
        mock_esi_character.client.Mail.get_characters_character_id_mail_lists.side_effect = (
            exception
        )
        mock_esi_sections.client.Mail.get_characters_character_id_mail_lists.side_effect = (
            exception
        )
        # when
        with self.assertRaises(HTTPError):
            tasks.update_character_mails(self.character_1001.pk, True)

        # then
        status: CharacterUpdateStatus = self.character_1001.update_status_set.get(
            section=Character.UpdateSection.MAILS
        )
        self.assertFalse(status.is_success)
        self.assertEqual(status.error_message, "HTTPBadGateway: 502 Test exception")
        self.assertTrue(status.run_started_at)
        self.assertTrue(status.run_finished_at)
        self.assertIsNone(status.update_started_at)
        self.assertIsNone(status.update_finished_at)


@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(TASKS_PATH + ".Location.objects.structure_update_or_create_esi", spec=True)
class TestUpdateStructureEsi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)
        cls.token = (
            cls.character.eve_character.character_ownership.user.token_set.first()
        )

    def test_normal(self, mock_structure_update_or_create_esi):
        """When ESI status is ok, then create MailEntity"""
        mock_structure_update_or_create_esi.return_value = None
        try:
            tasks.update_structure_esi(id=1_000_000_000_001, token_pk=self.token.pk)
        except Exception as ex:
            self.fail(f"Unexpected exception occurred: {ex}")

    def test_invalid_token(self, mock_structure_update_or_create_esi):
        """When called with invalid token, raise exception"""
        mock_structure_update_or_create_esi.side_effect = EsiOffline

        with self.assertRaises(Token.DoesNotExist):
            tasks.update_structure_esi(
                id=1_000_000_000_001, token_pk=generate_invalid_pk(Token)
            )

    def test_esi_status_1(self, mock_structure_update_or_create_esi):
        """When ESI is offline, then retry"""
        mock_structure_update_or_create_esi.side_effect = EsiOffline

        with self.assertRaises(CeleryRetry):
            tasks.update_structure_esi(id=1_000_000_000_001, token_pk=self.token.pk)

    def test_esi_status_2(self, mock_structure_update_or_create_esi):
        """When ESI error limit reached, then retry"""
        mock_structure_update_or_create_esi.side_effect = EsiErrorLimitExceeded(5)

        with self.assertRaises(CeleryRetry):
            tasks.update_structure_esi(id=1_000_000_000_001, token_pk=self.token.pk)


@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@patch(TASKS_PATH + ".MailEntity.objects.update_or_create_esi", spec=True)
class TestUpdateMailEntityEsi(TestCase):
    def test_normal(self, mock_update_or_create_esi):
        """When ESI status is ok, then create MailEntity"""
        mock_update_or_create_esi.return_value = None
        try:
            tasks.update_mail_entity_esi(1001)
        except Exception:
            self.fail("Unexpected exception occurred")

    def test_esi_status_1(self, mock_update_or_create_esi):
        """When ESI is offline, then abort"""
        mock_update_or_create_esi.side_effect = EsiOffline

        with self.assertRaises(EsiOffline):
            tasks.update_mail_entity_esi(1001)

    def test_esi_status_2(self, mock_update_or_create_esi):
        """When ESI error limit reached, then abort"""
        mock_update_or_create_esi.side_effect = EsiErrorLimitExceeded(5)

        with self.assertRaises(EsiErrorLimitExceeded):
            tasks.update_mail_entity_esi(1001)


@patch(MANAGERS_PATH + ".general.fetch_esi_status", lambda: EsiStatus(True, 99, 60))
@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestUpdateCharactersDoctrines(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        reset_celery_once_locks()

    @patch(MODELS_PATH + ".characters.Character.update_skill_sets")
    def test_normal(self, mock_update_skill_sets):
        # given
        mock_update_skill_sets.return_value = UpdateSectionResult(
            is_changed=True, is_updated=True
        )
        create_memberaudit_character(1001)

        # when
        tasks.update_characters_skill_checks()

        # then
        self.assertTrue(mock_update_skill_sets.called)


class TestDeleteCharacters(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        Character.objects.all().delete()

    def test_should_delete_a_character(self):
        # given
        character_1001 = create_memberaudit_character(1001)
        character_1002 = create_memberaudit_character(1002)
        # when
        tasks.delete_objects("Character", [character_1001.pk, character_1002.pk])
        # then
        self.assertFalse(Character.objects.exists())

    def test_should_raise_error_when_model_not_found(self):
        # when/then
        with self.assertRaises(LookupError):
            tasks.delete_objects("MyUnknownMOdel", [1])


@override_settings(
    CELERY_ALWAYS_EAGER=True,
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    APP_UTILS_OBJECT_CACHE_DISABLED=True,
)
class TestExportData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)
        reset_celery_once_locks()

    @patch(TASKS_PATH + ".data_exporters.export_topic_to_archive", spec=True)
    def test_should_export_all_topics(self, mock_export_topic_to_file):
        # when
        tasks.export_data()
        # then
        called_topics = [
            call[1]["topic"] for call in mock_export_topic_to_file.call_args_list
        ]
        self.assertEqual(len(called_topics), 3)
        self.assertSetEqual(
            set(called_topics), {"contract", "contract-item", "wallet-journal"}
        )

    @patch(TASKS_PATH + ".data_exporters.export_topic_to_archive", spec=True)
    def test_should_export_wallet_journal(self, mock_export_topic_to_file):
        # given
        user = self.character.user
        # when
        tasks.export_data_for_topic(topic="abc", user_pk=user.pk)
        # then
        self.assertTrue(mock_export_topic_to_file.called)
        _, kwargs = mock_export_topic_to_file.call_args
        self.assertEqual(kwargs["topic"], "abc")


class TestUpdateComplianceGroupDesignations(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()

    @patch(TASKS_PATH + ".ComplianceGroupDesignation.objects.update_user", spec=True)
    def test_should_update_for_user(self, mock_update_user):
        # given
        user, _ = create_user_from_evecharacter(
            1001,
            permissions=["memberaudit.basic_access"],
            scopes=Character.get_esi_scopes(),
        )
        # when
        tasks.update_compliance_groups_for_user(user.pk)
        # then
        self.assertTrue(mock_update_user.called)


@patch(TASKS_PATH + ".update_character", spec=True)
class TestUpdateAllCharacters(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()

    def test_should_update_all_enabled_characters(self, mock_update_character):
        # given
        character_1001 = create_memberaudit_character(1001)
        character_1002 = create_memberaudit_character(1002)
        character_1003 = create_memberaudit_character(1003)
        character_1003.is_disabled = True
        character_1003.save()
        # when
        tasks.update_all_characters()
        # then
        self.assertTrue(mock_update_character.apply_async.called)
        called_pks = {
            o[1]["kwargs"]["character_pk"]
            for o in mock_update_character.apply_async.call_args_list
        }
        self.assertSetEqual(called_pks, {character_1001.pk, character_1002.pk})

    def test_should_disable_orphaned_characters(self, mock_update_character):
        # given
        character_1001 = create_memberaudit_character(1001)
        eve_character_1002 = EveCharacter.objects.get(character_id=1002)
        character_1002 = create_character(eve_character_1002)
        # when
        tasks.update_all_characters()
        # then
        character_1001.refresh_from_db()
        self.assertFalse(character_1001.is_disabled)
        character_1002.refresh_from_db()
        self.assertTrue(character_1002.is_disabled)


@patch(TASKS_PATH + ".EveEntity.objects.update_from_esi_by_id", spec=True)
class TestUpdateUnresolvedEveEntities(TestCase):
    def test_should_not_attempt_to_update_when_no_unresolved_entities(
        self, mock_update_from_esi_by_id
    ):
        # given
        create_eve_entity(id=1, name="alpha")
        # when
        tasks.update_unresolved_eve_entities()
        # then
        self.assertFalse(mock_update_from_esi_by_id.called)

    def test_should_update_unresolved_entities(self, mock_update_from_esi_by_id):
        # given
        create_eve_entity(id=1)
        # when
        tasks.update_unresolved_eve_entities()
        # then
        self.assertTrue(mock_update_from_esi_by_id.called)
        args, _ = mock_update_from_esi_by_id.call_args
        self.assertEqual(list(args[0]), [1])


@patch(TASKS_PATH + ".check_character_consistency", spec=True)
class TestCheckCharacterConsistency(TestCase):
    def test_should_run_checks(self, mock_check_character_consistency):
        # given
        load_entities()
        character = create_memberaudit_character(1001)
        # when
        tasks.check_character_consistency(character.pk)
        # then
        self.assertTrue(mock_check_character_consistency.called)


class TestUpdateMarketPrices(TestCase):
    @patch(TASKS_PATH + ".EveMarketPrice.objects.update_from_esi", spec=True)
    def test_update_market_prices(self, mock_update_from_esi):
        tasks.update_market_prices()
        self.assertTrue(mock_update_from_esi.called)



"""Factories for creating test objects with defaults."""

import datetime as dt
import random
from itertools import count
from pathlib import Path
from typing import Iterable

from django.contrib.auth.models import Group, User
from django.utils.timezone import now
from eveuniverse.models import (
    EveEntity,
    EveMarketPrice,
    EvePlanet,
    EveSolarSystem,
    EveType,
)

from allianceauth.authentication.models import State
from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_authgroup

from memberaudit.core.fittings import Fitting, Item, Module
from memberaudit.core.skill_plans import SkillPlan
from memberaudit.core.skills import Skill
from memberaudit.models import (
    Character,
    CharacterAsset,
    CharacterAttributes,
    CharacterCloneInfo,
    CharacterContact,
    CharacterContactLabel,
    CharacterContract,
    CharacterContractBid,
    CharacterContractItem,
    CharacterCorporationHistory,
    CharacterDetails,
    CharacterFwStats,
    CharacterImplant,
    CharacterJumpClone,
    CharacterJumpCloneImplant,
    CharacterLocation,
    CharacterLoyaltyEntry,
    CharacterMail,
    CharacterMailLabel,
    CharacterMiningLedgerEntry,
    CharacterOnlineStatus,
    CharacterPlanet,
    CharacterRole,
    CharacterShip,
    CharacterSkill,
    CharacterSkillqueueEntry,
    CharacterSkillSetCheck,
    CharacterStanding,
    CharacterTitle,
    CharacterUpdateStatus,
    CharacterWalletJournalEntry,
    CharacterWalletTransaction,
    ComplianceGroupDesignation,
    Location,
    MailEntity,
    SkillSet,
    SkillSetGroup,
    SkillSetSkill,
)

from .constants import (
    EveCategoryId,
    EveFactionId,
    EveSolarSystemId,
    EveStationId,
    EveTypeId,
)


def create_character(eve_character: EveCharacter, **kwargs) -> Character:
    params = {"eve_character": eve_character}
    params.update(kwargs)
    obj = Character(**params)
    obj.save(ignore_cache=True)
    return obj


def create_character_from_user(user: User, **kwargs):
    """Create new Character object from user. The user needs to have a main character.

    This factory is designed to work with both the old and new variant of Character
    introduced in version 2.
    """
    eve_character = user.profile.main_character
    if not eve_character:
        raise ValueError("User needs to have a main character.")

    kwargs.update({"eve_character": eve_character})
    return create_character(**kwargs)


def build_character_asset(character: Character, **kwargs) -> CharacterAsset:
    item_id = kwargs.get("item_id") or next_number("asset_item_id") + 1_200_000_000_000
    params = {
        "character": character,
        "item_id": item_id,
        "is_singleton": False,
        "quantity": 1,
        "location_flag": "Hangar",
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.VELDSPAR)
    if params["is_singleton"] and not params.get("name"):
        params["name"] = (f"Generated asset #{item_id}",)
    return CharacterAsset(**params)


def create_character_asset(character: Character, **kwargs) -> CharacterAsset:
    obj = build_character_asset(character, **kwargs)
    obj.save()
    return obj


def create_character_attributes(character: Character, **kwargs) -> CharacterAttributes:
    params = {
        "character": character,
        "bonus_remaps": 4,
        "charisma": 16,
        "intelligence": 17,
        "memory": 18,
        "perception": 19,
        "willpower": 20,
    }
    params.update(kwargs)
    return CharacterAttributes.objects.create(**params)


def create_character_clone_info(character: Character, **kwargs) -> CharacterCloneInfo:
    params = {
        "character": character,
        "last_clone_jump_date": now() - dt.timedelta(days=7),
        "last_station_change_date": now() - dt.timedelta(days=90),
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, home_location_id=1000000000001)
    return CharacterCloneInfo.objects.create(**params)


def create_character_contact(
    character: Character, eve_entity: EveEntity, **kwargs
) -> CharacterContact:
    params = {
        "character": character,
        "eve_entity": eve_entity,
        "is_blocked": False,
        "is_watched": False,
        "standing": 0.0,
    }
    params.update(kwargs)
    return CharacterContact.objects.create(**params)


def create_character_contact_label(
    character: Character, **kwargs
) -> CharacterContactLabel:
    label_id = kwargs.get("label_id") or next_number("character_contact_label_id") + 100
    params = {
        "character": character,
        "label_id": label_id,
        "name": f"Test Label #{label_id}",
    }
    params.update(kwargs)
    return CharacterContactLabel.objects.create(**params)


def create_character_contract(character: Character, **kwargs) -> CharacterContract:
    date_issued = kwargs.get("date_issued") or now()
    contract_id = kwargs.get("contract_id") or next_number("contract_id") + 190_000_000
    params = {
        "character": character,
        "contract_id": contract_id,
        "availability": CharacterContract.AVAILABILITY_PERSONAL,
        "contract_type": CharacterContract.TYPE_ITEM_EXCHANGE,
        "date_issued": date_issued,
        "date_expired": date_issued + dt.timedelta(days=3),
        "for_corporation": False,
        "status": CharacterContract.STATUS_OUTSTANDING,
        "title": f"Test Contract #{contract_id}",
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, assignee_id=1002)
    if "issuer_id" not in params and "issuer" not in kwargs:
        params["issuer_id"] = 1001
        params["issuer_corporation_id"] = 2001
    return CharacterContract.objects.create(**params)


def create_character_contract_courier(
    character: Character, **kwargs
) -> CharacterContract:
    start_location = kwargs.get("start_location") or create_location()
    end_location = kwargs.get("end_location") or create_location()
    params = {
        "character": character,
        "availability": CharacterContract.AVAILABILITY_PERSONAL,
        "contract_type": CharacterContract.TYPE_COURIER,
        "days_to_complete": 7,
        "end_location": end_location,
        "reward": 100_000_000,
        "start_location": start_location,
        "status": CharacterContract.STATUS_OUTSTANDING,
        "volume": 250_000,
    }
    params.update(kwargs)
    return create_character_contract(**params)


def create_character_contract_item(
    contract: CharacterContract, **kwargs
) -> CharacterContractItem:
    record_id = kwargs.get("record_id") or next_number("contract_item_record_id")
    params = {
        "contract": contract,
        "record_id": record_id,
        "is_included": True,
        "is_singleton": False,
        "quantity": 1,
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.MERLIN)
    return CharacterContractItem.objects.create(**params)


def create_character_contract_bid(
    contract: CharacterContract, bidder: EveEntity, **kwargs
) -> CharacterContractBid:
    bid_id = kwargs.get("bid_id") or next_number("contract_item_bid_id")
    params = {
        "contract": contract,
        "bid_id": bid_id,
        "amount": random.randint(1_000_000, 10_000_000_000),
        "bidder": bidder,
        "date_bid": now(),
    }
    params.update(kwargs)
    return CharacterContractBid.objects.create(**params)


def create_character_corporation_history(
    character: Character, **kwargs
) -> CharacterCorporationHistory:
    record_id = kwargs.get("record_id") or next_number(
        "create_character_details_record_id"
    )
    weeks = max(0, 1000 - record_id * 20)
    params = {
        "character": character,
        "record_id": record_id,
        "start_date": now() - dt.timedelta(weeks=weeks),
    }
    params.update(kwargs)
    return CharacterCorporationHistory.objects.create(**params)


def create_character_details(character: Character, **kwargs) -> CharacterDetails:
    params = {
        "character": character,
        "birthday": now() - dt.timedelta(weeks=200),
        "name": character.eve_character.character_name,
    }
    params.update(kwargs)
    _set_missing_foreign_keys(
        params,
        alliance_id=character.eve_character.alliance_id,
        corporation_id=character.eve_character.corporation_id,
        eve_bloodline_id=1,
        eve_race_id=1,
    )
    return CharacterDetails.objects.create(**params)


def create_character_fw_stats(character: Character, **kwargs) -> CharacterFwStats:
    current_rank = random.randint(1, 5)
    highest_rank = max(current_rank, random.randint(1, 5))
    kills_yesterday = random.randint(1, 100)
    kills_last_week = max(kills_yesterday, random.randint(1, 700))
    kills_total = max(kills_last_week, random.randint(1, 5000))
    victory_points_yesterday = random.randint(1, 1000)
    victory_points_last_week = max(victory_points_yesterday, random.randint(1, 7000))
    victory_points_total = max(victory_points_last_week, random.randint(1, 50000))
    enlisted_on = now() - dt.timedelta(
        days=random.randint(1, 180), hours=random.randint(1, 24)
    )
    params = {
        "character": character,
        "current_rank": current_rank,
        "enlisted_on": enlisted_on,
        "highest_rank": highest_rank,
        "kills_last_week": kills_last_week,
        "kills_total": kills_total,
        "kills_yesterday": kills_yesterday,
        "victory_points_last_week": victory_points_last_week,
        "victory_points_total": victory_points_total,
        "victory_points_yesterday": victory_points_yesterday,
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, faction_id=EveFactionId.CALDARI_STATE)
    return CharacterFwStats.objects.create(**params)


def create_character_implant(character: Character, **kwargs) -> CharacterImplant:
    params = {"character": character}
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.HIGH_GRADE_SNAKE_ALPHA)
    return CharacterImplant.objects.create(**params)


def create_character_jump_clone(character: Character, **kwargs) -> CharacterJumpClone:
    jump_clone_id = kwargs.get("jump_clone_id") or next_number(
        "create_character_implant_jump_clone_id"
    )
    params = {"character": character, "jump_clone_id": jump_clone_id}
    params.update(kwargs)
    _set_missing_foreign_keys(params, location_id=EveStationId.JITA_44)
    return CharacterJumpClone.objects.create(**params)


def create_character_jump_clone_implant(
    jump_clone: CharacterJumpClone, **kwargs
) -> CharacterJumpCloneImplant:
    params = {"jump_clone": jump_clone}
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.HIGH_GRADE_SNAKE_ALPHA)
    return CharacterJumpCloneImplant.objects.create(**params)


def create_character_location(character: Character, **kwargs) -> CharacterLocation:
    params = {"character": character}
    params.update(kwargs)
    location = kwargs.get("location")
    if location and location.eve_solar_system:
        params["eve_solar_system"] = location.eve_solar_system
    else:
        _set_missing_foreign_keys(params, eve_solar_system_id=EveSolarSystemId.AMAMAKE)
    return CharacterLocation.objects.create(**params)


def create_character_loyalty_entry(
    character: Character, **kwargs
) -> CharacterLoyaltyEntry:
    params = {
        "character": character,
        "loyalty_points": random.randint(100, 1_000_000),
    }
    params.update(kwargs)
    return CharacterLoyaltyEntry.objects.create(**params)


def create_character_mail(
    character: Character,
    recipients: Iterable[MailEntity] = None,
    labels: Iterable[CharacterMailLabel] = None,
    **kwargs,
) -> CharacterMail:
    timestamp = kwargs.get("timestamp") or now()
    mail_id = kwargs.get("mail_id") or next_number("mail_id")
    params = {
        "character": character,
        "mail_id": mail_id,
        "subject": f"Test Subject #{mail_id}",
        "body": f"Test Body #{mail_id}",
        "timestamp": timestamp,
    }
    if "sender" not in kwargs and "sender_id" not in kwargs:
        params["sender"] = create_mail_entity_from_eve_entity(id=1002)
    params.update(kwargs)
    obj = CharacterMail.objects.create(**params)
    if not recipients:
        character_id = character.eve_character.character_id
        recipients = [create_mail_entity_from_eve_entity(id=character_id)]
    obj.recipients.add(*recipients)
    if labels:
        obj.labels.add(*labels)
    return obj


def create_character_mail_label(character: Character, **kwargs) -> CharacterMailLabel:
    label_id = kwargs.get("label_id") or next_number("mail_label_id")
    params = {
        "character": character,
        "label_id": label_id,
        "name": f"Label #{label_id}",
    }
    params.update(kwargs)
    return CharacterMailLabel.objects.create(**params)


def create_character_mining_ledger_entry(
    character: Character, **kwargs
) -> CharacterMiningLedgerEntry:
    solar_system_ids = EveSolarSystem.objects.values_list("id", flat=True)
    ore_type_ids = EveType.objects.filter(
        eve_group__eve_category_id=EveCategoryId.ASTEROID, published=True
    ).values_list("id", flat=True)
    params = {
        "character": character,
        "date": (now() - dt.timedelta(days=random.randint(0, 300))).date(),
        "quantity": random.randint(10_000, 50_000),
        "eve_solar_system": EveSolarSystem.objects.get(
            id=random.choice(solar_system_ids)
        ),
        "eve_type": EveType.objects.get(id=random.choice(ore_type_ids)),
    }
    params.update(kwargs)
    return CharacterMiningLedgerEntry.objects.create(**params)


def create_character_online_status(
    character: Character, **kwargs
) -> CharacterOnlineStatus:
    params = {
        "character": character,
        "last_login": now() - dt.timedelta(days=1),
        "last_logout": now() - dt.timedelta(days=1) + dt.timedelta(hours=4),
        "logins": 42,
    }
    params.update(kwargs)
    return CharacterOnlineStatus.objects.create(**params)


def create_character_planet(character: Character, **kwargs) -> CharacterPlanet:
    all_planets = set(EvePlanet.objects.values_list("id", flat=True))
    colonized_planets = set(
        CharacterPlanet.objects.values_list("eve_planet_id", flat=True)
    )
    available_planets = all_planets - colonized_planets
    if not available_planets:
        raise RuntimeError("No free planet to colonize")
    params = {
        "character": character,
        "last_update_at": now() - dt.timedelta(days=random.randint(0, 300)),
        "num_pins": random.randint(1, 10),
        "eve_planet": EvePlanet.objects.get(id=random.choice(list(available_planets))),
        "upgrade_level": random.randint(0, 5),
    }
    params.update(kwargs)
    return CharacterPlanet.objects.create(**params)


def create_character_role(character: Character, **kwargs) -> CharacterRole:
    params = {
        "character": character,
        "role": CharacterRole.Role.DIRECTOR,
        "location": CharacterRole.Location.UNIVERSAL,
    }
    params.update(kwargs)
    return CharacterRole.objects.create(**params)


def create_character_ship(character: Character, **kwargs) -> CharacterShip:
    item_id = kwargs.get("item_id") or next_number("asset_item_id")
    params = {"character": character, "item_id": item_id, "name": "My sweet ride"}
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.MERLIN)
    return CharacterShip.objects.create(**params)


def create_character_skill(character: Character, **kwargs) -> CharacterSkill:
    params = {
        "character": character,
        "active_skill_level": 3,
        "skillpoints_in_skill": 1000,
        "trained_skill_level": 3,
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.AMARR_CARRIER)
    return CharacterSkill.objects.create(**params)


def create_character_skillqueue_entry(
    character: Character, **kwargs
) -> CharacterSkillqueueEntry:
    queue_position = kwargs.get("queue_position") or next_number(
        "skillqueue_queue_position"
    )
    params = {
        "character": character,
        "finished_level": 3,
        "queue_position": queue_position,
        "level_end_sp": 512,
        "level_start_sp": 128,
        "start_date": now(),
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.AMARR_CARRIER)
    return CharacterSkillqueueEntry.objects.create(**params)


def create_character_skill_set_check(
    character: Character, skill_set: SkillSet, **kwargs
) -> CharacterSkillSetCheck:
    return CharacterSkillSetCheck.objects.create(
        character=character, skill_set=skill_set, **kwargs
    )


def create_character_standing(
    character: Character, eve_entity: EveEntity, **kwargs
) -> CharacterStanding:
    params = {
        "character": character,
        "eve_entity": eve_entity,
        "standing": 0.0,
    }
    params.update(kwargs)
    return CharacterStanding.objects.create(**params)


def create_character_title(character: Character, **kwargs) -> CharacterRole:
    title_id = kwargs.get("title_id") or next_number("title_id")
    params = {
        "character": character,
        "name": f"Dummy title #{title_id}",
        "title_id": title_id,
    }
    params.update(kwargs)
    return CharacterTitle.objects.create(**params)


def create_character_update_status(
    character: Character, **kwargs
) -> CharacterUpdateStatus:
    params = {
        "character": character,
        "section": Character.UpdateSection.ASSETS,
        "is_success": True,
        "run_started_at": now() - dt.timedelta(minutes=5),
        "run_finished_at": now(),
    }
    params.update(kwargs)
    return CharacterUpdateStatus.objects.create(**params)


def create_character_wallet_journal_entry(
    character: Character, **kwargs
) -> CharacterWalletJournalEntry:
    entry_id = kwargs.get("entry_id") or next_number("wallet_journal_entry_id")
    params = {
        "character": character,
        "entry_id": entry_id,
        "amount": 1000000.0,
        "balance": 20000000.0,
        "ref_type": "player_donation",
        "context_id_type": CharacterWalletJournalEntry.CONTEXT_ID_TYPE_UNDEFINED,
        "date": now(),
        "description": "test description",
        "reason": "test reason",
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, first_party_id=1001, second_party_id=1002)
    return CharacterWalletJournalEntry.objects.create(**params)


def create_character_wallet_transaction(
    character: Character, **kwargs
) -> CharacterWalletTransaction:
    transaction_id = kwargs.get("transaction_id") or next_number(
        "create_character_wallet_transaction_transaction_id"
    )
    params = {
        "character": character,
        "transaction_id": transaction_id,
        "date": now(),
        "is_buy": True,
        "is_personal": True,
        "quantity": 1,
        "unit_price": float(random.randint(10_000, 100_000_000)) + random.random(),
    }
    params.update(kwargs)
    _set_missing_foreign_keys(
        params,
        client_id=1002,
        location_id=EveStationId.JITA_44,
        eve_type_id=EveTypeId.MERLIN,
    )
    return CharacterWalletTransaction.objects.create(**params)


def create_compliance_group(states: Iterable[State] = None, **kwargs) -> Group:
    group = create_authgroup(states, internal=True, **kwargs)
    create_compliance_group_designation(group)
    return group


def create_compliance_group_designation(
    group: Group, **kwargs
) -> ComplianceGroupDesignation:
    params = {"group": group}
    params.update(kwargs)
    return ComplianceGroupDesignation.objects.create(**params)


def create_fitting(**kwargs):
    """Requires eveuniverse to be loaded."""
    params = {
        "name": "Test fitting",
        "ship_type": EveType.objects.get(name="Tristan"),
        "high_slots": [
            Module(
                EveType.objects.get(name="125mm Gatling AutoCannon II"),
                charge_type=EveType.objects.get(name="EMP S"),
            ),
            None,
        ],
        "medium_slots": [Module(EveType.objects.get(name="Warp Disruptor II")), None],
        "low_slots": [
            Module(EveType.objects.get(name="Drone Damage Amplifier II")),
            None,
        ],
        "rig_slots": [
            Module(EveType.objects.get(name="Small EM Shield Reinforcer I")),
            None,
        ],
        "drone_bay": [Item(EveType.objects.get(name="Acolyte II"), quantity=5)],
        "cargo_bay": [Item(EveType.objects.get(name="EMP S"), quantity=3)],
    }
    params.update(kwargs)
    return Fitting(**params)


def create_fitting_text(file_name: str) -> str:
    testdata_folder = Path(__file__).parent / "fittings"
    fitting_file = testdata_folder / file_name
    with fitting_file.open("r") as file:
        return file.read()


def create_location(**kwargs) -> Location:
    location_id = kwargs.get("id") or next_number("location_id") + 1_700_000_000_000
    params = {
        "id": location_id,
        "name": f"Test Location #{location_id}",
    }
    params.update(kwargs)
    _set_missing_foreign_keys(
        params,
        eve_solar_system_id=EveSolarSystemId.AMAMAKE,
        eve_type_id=EveTypeId.ASTRAHUS,
    )
    return Location.objects.create(**params)


def create_location_eve_solar_system(**kwargs) -> Location:
    solar_system_id = kwargs.get("id") or EveSolarSystemId.AMAMAKE
    eve_solar_system = EveSolarSystem.objects.get(id=solar_system_id)
    params = {"name": eve_solar_system.name, "eve_type_id": EveTypeId.SOLAR_SYSTEM}
    params.update(kwargs)
    return create_location(**params)


def create_mail_entity(**kwargs) -> MailEntity:
    my_id = kwargs.get("id") or next_number("create_mail_entity_id") + 10_000_000
    params = {
        "id": my_id,
        "name": "",
        "category": MailEntity.Category.UNKNOWN,
    }
    params.update(kwargs)
    return MailEntity.objects.create(**params)


def create_mail_entity_from_eve_entity(id: int) -> MailEntity:
    obj, _ = MailEntity.objects.update_or_create_from_eve_entity_id(id=id)
    return obj


def create_mailing_list(**kwargs) -> MailEntity:
    my_id = kwargs.get("id") or next_number("create_mail_entity_id") + 20_000_000
    params = {
        "id": my_id,
        "name": f"Mailing List #{my_id}",
        "category": MailEntity.Category.MAILING_LIST,
    }
    params.update(kwargs)
    return create_mail_entity(**params)


def create_skill(**kwargs) -> Skill:
    params = {}
    if "eve_type" not in kwargs:
        params["eve_type"] = (
            EveType.objects.filter(
                eve_group__eve_category_id=EveCategoryId.SKILL, published=True
            )
            .order_by("?")
            .first()
        )
    if "level" not in kwargs:
        params["level"] = random.randint(1, 5)
    params.update(kwargs)
    return Skill(**params)


def create_skill_plan(**kwargs) -> SkillPlan:
    my_id = next_number("skill_plan_id")
    params = {"name": f"Test Skill Plan {my_id}"}
    if "skills" not in kwargs:
        params["skills"] = [create_skill() for _ in range(random.randint(1, 5))]
    params.update(kwargs)
    return SkillPlan(**params)


def create_skill_set(**kwargs) -> SkillSet:
    my_id = next_number("skill_set_id")
    params = {
        "name": f"Merlin #{my_id}",
        "description": "Generated skill set",
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, ship_type_id=EveTypeId.MERLIN)
    return SkillSet.objects.create(**params)


def create_skill_set_group(**kwargs) -> SkillSetGroup:
    my_id = next_number("skill_set_group_id")
    params = {
        "name": f"Test Group #{my_id}",
        "description": "Generated skill set group",
    }
    params.update(kwargs)
    return SkillSetGroup.objects.create(**params)


def create_skill_set_skill(skill_set, **kwargs) -> SkillSetSkill:
    params = {
        "skill_set": skill_set,
        "required_level": 1,
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.AMARR_CARRIER)
    return SkillSetSkill.objects.create(**params)


def create_eve_market_price(**kwargs) -> EveMarketPrice:
    average_price = (
        kwargs.get("average_price")
        or float(random.randint(10_000, 100_000_000)) + random.random()
    )
    adjusted_price = kwargs.get("adjusted_price") or average_price
    params = {
        "average_price": average_price,
        "adjusted_price": adjusted_price,
    }
    params.update(kwargs)
    _set_missing_foreign_keys(params, eve_type_id=EveTypeId.VELDSPAR)
    return EveMarketPrice.objects.create(**params)


def next_number(key=None, offset=None) -> int:
    if key is None:
        key = "_general"
    try:
        return next_number._counter[key].__next__()
    except AttributeError:
        next_number._counter = {}
    except KeyError:
        pass
    next_number._counter[key] = count(start=1)
    number = next_number._counter[key].__next__()
    result = number + offset if offset else number
    return result


def _set_missing_foreign_keys(params: dict, **kwargs):
    """Set foreign keys in dict when they and their variant are not present."""
    for key_id, value in kwargs.items():
        key = key_id.replace("_id", "")
        if key not in params and key_id not in params:
            params[key_id] = int(value)



import datetime as dt
from unittest.mock import patch

from django.db import IntegrityError
from django.test import TestCase, override_settings
from django.utils.dateparse import parse_datetime
from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveType

from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import NoSocketsTestCase

from memberaudit.models import (
    CharacterAsset,
    CharacterAttributes,
    CharacterContactLabel,
    CharacterContract,
    CharacterContractBid,
    CharacterContractItem,
    Location,
)
from memberaudit.tests.testdata.constants import EveStationId, EveTypeId
from memberaudit.tests.testdata.esi_client_stub import esi_client_stub
from memberaudit.tests.testdata.factories import (
    build_character_asset,
    create_character_asset,
    create_character_attributes,
    create_character_contact,
    create_character_contact_label,
    create_character_contract,
    create_character_contract_bid,
    create_character_from_user,
    create_character_skillqueue_entry,
    create_eve_market_price,
    create_location,
)
from memberaudit.tests.testdata.load_entities import load_entities
from memberaudit.tests.testdata.load_eveuniverse import load_eveuniverse
from memberaudit.tests.testdata.load_locations import load_locations
from memberaudit.tests.utils import (
    create_memberaudit_character,
    create_user_from_evecharacter_with_access,
)

MODULE_PATH = "memberaudit.managers.character_sections_1"


class TestCharacterSkillQueue(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)
        cls.skill_type_1 = EveType.objects.get(id=24311)
        cls.skill_type_2 = EveType.objects.get(id=24312)

    def test_is_active_1(self):
        """when training is active and skill is in first position then return True"""
        entry = create_character_skillqueue_entry(
            character=self.character,
            eve_type=self.skill_type_1,
            finish_date=now() + dt.timedelta(days=3),
            queue_position=0,
            start_date=now() - dt.timedelta(days=1),
        )
        self.assertTrue(entry.is_active)

    def test_is_active_2(self):
        """when training is active and skill is not in first position then return False"""
        entry = create_character_skillqueue_entry(
            character=self.character,
            eve_type=self.skill_type_1,
            finish_date=now() + dt.timedelta(days=3),
            finished_level=5,
            queue_position=1,
            start_date=now() - dt.timedelta(days=1),
        )
        self.assertFalse(entry.is_active)

    def test_is_active_3(self):
        """when training is not active and skill is in first position then return False"""
        entry = create_character_skillqueue_entry(
            character=self.character,
            eve_type=self.skill_type_1,
            finished_level=5,
            queue_position=0,
        )
        self.assertFalse(entry.is_active)


class TestCharacterAssetManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character = create_memberaudit_character(1001)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.merlin = EveType.objects.get(id=603)

    def test_can_calculate_pricing(self):
        create_character_asset(
            character=self.character, eve_type=self.merlin, quantity=5
        )
        create_eve_market_price(eve_type=self.merlin, average_price=500000)
        asset = CharacterAsset.objects.annotate_pricing().first()
        self.assertEqual(asset.price, 500000)
        self.assertEqual(asset.total, 2500000)

    def test_does_not_price_blueprint_copies(self):
        create_character_asset(
            character=self.character,
            eve_type=self.merlin,
            is_blueprint_copy=True,
            quantity=1,
        )
        create_eve_market_price(eve_type=self.merlin, average_price=500000)
        asset = CharacterAsset.objects.annotate_pricing().first()
        self.assertIsNone(asset.price)
        self.assertIsNone(asset.total)


class TestCharacterAssetManagerBulkCreate2(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character = create_memberaudit_character(1001)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.merlin = EveType.objects.get(id=603)

    def test_should_create_assets_in_bulk(self):
        # given
        objs = [build_character_asset(self.character) for _ in range(5)]

        # when
        new_objs = CharacterAsset.objects.bulk_create_with_fallback(objs)

        # then
        expected_ids = _extract_item_ids(objs)
        existing_item_ids = set(self.character.assets.values_list("item_id", flat=True))
        self.assertSetEqual(existing_item_ids, expected_ids)
        self.assertSetEqual(_extract_item_ids(new_objs), expected_ids)

    def test_should_create_all_assets_and_ignore_the_problem_obj(self):
        # given
        objs = [build_character_asset(self.character) for _ in range(5)]
        problem_item_id = objs[3].item_id

        def my_save(obj: CharacterAsset, *args, **kwargs):
            if int(obj.item_id) == problem_item_id:
                raise IntegrityError("Test exception")
            super(CharacterAsset, obj).save(*args, **kwargs)

        # when
        with patch.object(
            CharacterAsset.objects, "bulk_create"
        ) as mock_bulk_create, patch(
            "memberaudit.models.character_sections_1.CharacterAsset.save", my_save
        ):
            mock_bulk_create.side_effect = IntegrityError("Test exception")

            new_objs = CharacterAsset.objects.bulk_create_with_fallback(objs)

        # then
        expected_ids = _extract_item_ids(objs) - {problem_item_id}
        self.assertSetEqual(_extract_item_ids(new_objs), expected_ids)

        existing_item_ids = set(self.character.assets.values_list("item_id", flat=True))
        self.assertSetEqual(existing_item_ids, expected_ids)


def _extract_item_ids(objs) -> set:
    return {obj.item_id for obj in objs}


@patch(MODULE_PATH + ".esi")
class TestCharacterAssetsFetchFromEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_can_fetch_new_assets(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        result = CharacterAsset.objects.fetch_from_esi(self.character)
        # then
        self.assertTrue(result.is_changed)
        asset_data = {asset["item_id"]: asset for asset in result.data}
        self.assertSetEqual(
            set(asset_data.keys()),
            {
                1100000000001,
                1100000000002,
                1100000000003,
                1100000000004,
                1100000000005,
                1100000000006,
                1100000000007,
                1100000000008,
            },
        )
        obj = asset_data[1100000000001]
        self.assertEqual(obj["name"], "Parent Item 1")
        self.assertTrue(obj["is_blueprint_copy"])
        self.assertTrue(obj["is_singleton"])
        self.assertEqual(obj["location_flag"], "Hangar")
        self.assertEqual(obj["location_id"], EveStationId.JITA_44.value)
        self.assertEqual(obj["quantity"], 1)
        self.assertEqual(obj["type_id"], EveTypeId.CHARON.value)

    def test_should_always_return_assets_when_forced(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        CharacterAsset.objects.fetch_from_esi(self.character)
        # when
        result = CharacterAsset.objects.fetch_from_esi(
            self.character, force_update=True
        )
        # then
        self.assertIsNotNone(result.data)
        self.assertFalse(result.is_changed)


@patch("memberaudit.models.Location.objects.create_missing_esi", spec=True)
@patch(MODULE_PATH + ".EveType.objects.bulk_get_or_create_esi", spec=True)
class TestCharacterAssetsPreloadObjects(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        cls.user, _ = create_user_from_evecharacter_with_access(1001)

    def test_do_nothing_when_asset_list_is_empty(
        self, mock_eve_entity_create, mock_preload_locations
    ):
        # given
        character = create_character_from_user(self.user)
        asset_list = []

        # when
        result = character.assets_preload_objects(asset_list)

        # then
        self.assertFalse(result.is_updated)
        self.assertFalse(mock_eve_entity_create.called)
        self.assertFalse(mock_preload_locations.called)

    def test_fetch_missing_eve_entity_objects_and_locations(
        self, mock_eve_entity_create, mock_preload_locations
    ):
        # given
        character = create_character_from_user(self.user)
        asset_list = [
            {"item_id": 1, "type_id": 3, "location_id": 420},
            {"item_id": 2, "type_id": 4, "location_id": 421},
        ]
        # when
        result = character.assets_preload_objects(asset_list)

        # then
        self.assertTrue(result.is_updated)
        self.assertTrue(mock_eve_entity_create.called)
        _, kwargs = mock_eve_entity_create.call_args
        self.assertEqual(set(kwargs["ids"]), {3, 4})
        self.assertTrue(mock_preload_locations.called)
        _, kwargs = mock_preload_locations.call_args
        self.assertEqual(kwargs["location_ids"], {420, 421})

    def test_fetch_missing_eve_entity_objects_only(
        self, mock_eve_entity_create, mock_preload_locations
    ):
        # given
        create_location(id=420)
        create_location(id=421)
        character = create_character_from_user(self.user)
        asset_list = [
            {"item_id": 1, "type_id": 3, "location_id": 420},
            {"item_id": 2, "type_id": 4, "location_id": 421},
        ]
        # when
        result = character.assets_preload_objects(asset_list)

        # then
        self.assertTrue(result.is_updated)
        self.assertTrue(mock_eve_entity_create.called)
        _, kwargs = mock_eve_entity_create.call_args
        self.assertEqual(set(kwargs["ids"]), {3, 4})
        self.assertFalse(mock_preload_locations.called)


@patch(MODULE_PATH + ".esi")
class TestCharacterAttributesManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_can_create_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        result = CharacterAttributes.objects.update_or_create_esi(self.character)
        # then
        self.assertTrue(result.is_changed)
        self.assertEqual(
            self.character.attributes.accrued_remap_cooldown_date,
            parse_datetime("2016-10-24T09:00:00Z"),
        )
        self.assertEqual(
            self.character.attributes.last_remap_date,
            parse_datetime("2016-10-24T09:00:00Z"),
        )
        self.assertEqual(self.character.attributes.charisma, 16)
        self.assertEqual(self.character.attributes.intelligence, 17)
        self.assertEqual(self.character.attributes.memory, 18)
        self.assertEqual(self.character.attributes.perception, 19)
        self.assertEqual(self.character.attributes.willpower, 20)

    def test_can_update_existing_attributes(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        create_character_attributes(
            character=self.character,
            accrued_remap_cooldown_date=None,
            last_remap_date=None,
            bonus_remaps=4,
            charisma=102,
            intelligence=103,
            memory=104,
            perception=105,
            willpower=106,
        )
        # when
        result = CharacterAttributes.objects.update_or_create_esi(self.character)

        # then
        self.assertTrue(result.is_changed)
        self.character.attributes.refresh_from_db()
        self.assertEqual(
            self.character.attributes.accrued_remap_cooldown_date,
            parse_datetime("2016-10-24T09:00:00Z"),
        )
        self.assertEqual(
            self.character.attributes.last_remap_date,
            parse_datetime("2016-10-24T09:00:00Z"),
        )
        self.assertEqual(self.character.attributes.charisma, 16)
        self.assertEqual(self.character.attributes.intelligence, 17)
        self.assertEqual(self.character.attributes.memory, 18)
        self.assertEqual(self.character.attributes.perception, 19)
        self.assertEqual(self.character.attributes.willpower, 20)


@patch(MODULE_PATH + ".esi")
class TestCharacterContactLabelManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_do_nothing_when_no_esi_data(self, mock_esi):
        # when
        endpoints = [
            EsiEndpoint(
                "Contacts",
                "get_characters_character_id_contacts_labels",
                "character_id",
                needs_token=True,
                data={"1001": []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # then
        self.assertEqual(CharacterContactLabel.objects.count(), 0)

    def test_should_create_labels_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub

        # when
        self.character.update_contact_labels()

        # then
        self.assertEqual(self.character.contact_labels.count(), 2)

        label = self.character.contact_labels.get(label_id=1)
        self.assertEqual(label.name, "friend")

        label = self.character.contact_labels.get(label_id=2)
        self.assertEqual(label.name, "pirate")

    def test_should_remove_obsolete_labels(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(character=self.character, label_id=99)

        # when
        self.character.update_contact_labels()

        # then
        self.assertSetEqual(self._current_label_ids(), {1, 2})

    def test_should_update_existing_label(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(
            character=self.character, label_id=1, name="update-me"
        )
        with patch(
            MODULE_PATH + ".CharacterContactLabelManager.bulk_update",
            wraps=CharacterContactLabel.objects.filter(
                character=self.character
            ).bulk_update,
        ) as mock_bulk_update:
            # when
            self.character.update_contact_labels()

            # then
            self.assertSetEqual(self._current_label_ids(), {1, 2})

            label = self.character.contact_labels.get(label_id=1)
            self.assertEqual(label.name, "friend")

            # then only the modified label was updated
            updated_obj_ids = {o.id for o in mock_bulk_update.call_args.kwargs["objs"]}
            self.assertSetEqual(updated_obj_ids, {label.id})

    def test_should_skip_update_when_esi_data_has_not_changed(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        self.character.update_contact_labels()
        label = self.character.contact_labels.get(label_id=1)
        label.name = "foe"
        label.save()

        # when
        self.character.update_contact_labels()

        # then
        self.assertEqual(self.character.contact_labels.count(), 2)
        label = self.character.contact_labels.get(label_id=1)
        self.assertEqual(label.name, "foe")

    def test_should_do_update_when_esi_data_has_not_changed_and_forced(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        self.character.update_contact_labels()
        label = self.character.contact_labels.get(label_id=1)
        label.name = "foe"
        label.save()

        # when
        self.character.update_contact_labels(force_update=True)

        self.assertEqual(self.character.contact_labels.count(), 2)
        label = self.character.contact_labels.get(label_id=1)
        self.assertEqual(label.name, "friend")

    def _current_label_ids(self):
        current_label_ids = {
            obj.label_id for obj in self.character.contact_labels.all()
        }
        return current_label_ids


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
class TestCharacterContactsManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_create_from_scratch(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(character=self.character, label_id=1)
        create_character_contact_label(character=self.character, label_id=2)

        # when
        result = self.character.update_contacts()

        # then
        self.assertTrue(result.is_changed)
        self.assertEqual(self.character.contacts.count(), 2)

        obj = self.character.contacts.get(eve_entity_id=1101)
        self.assertEqual(obj.eve_entity.category, EveEntity.CATEGORY_CHARACTER)
        self.assertFalse(obj.is_blocked)
        self.assertTrue(obj.is_watched)
        self.assertEqual(obj.standing, -10)
        self.assertEqual({x.label_id for x in obj.labels.all()}, {2})

        obj = self.character.contacts.get(eve_entity_id=2002)
        self.assertEqual(obj.eve_entity.category, EveEntity.CATEGORY_CORPORATION)
        self.assertFalse(obj.is_blocked)
        self.assertFalse(obj.is_watched)
        self.assertEqual(obj.standing, 5)
        self.assertEqual(obj.labels.count(), 0)

    def test_should_remove_obsolete_contacts(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(character=self.character, label_id=1)
        create_character_contact_label(character=self.character, label_id=2)
        create_character_contact(
            character=self.character,
            eve_entity=EveEntity.objects.get(id=3101),
            standing=-5,
        )

        # when
        result = self.character.update_contacts()

        # then
        self.assertTrue(result.is_changed)
        self.assertEqual(
            {x.eve_entity_id for x in self.character.contacts.all()}, {1101, 2002}
        )

    def test_should_update_existing_contracts(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(character=self.character, label_id=2)
        my_label = create_character_contact_label(character=self.character, label_id=1)
        my_contact = create_character_contact(
            character=self.character,
            eve_entity=EveEntity.objects.get(id=1101),
            is_blocked=True,
            is_watched=False,
            standing=-5,
        )
        my_contact.labels.add(my_label)

        # when
        result = self.character.update_contacts()

        # then
        self.assertTrue(result.is_changed)
        obj = self.character.contacts.get(eve_entity_id=1101)
        self.assertEqual(obj.eve_entity.category, EveEntity.CATEGORY_CHARACTER)
        self.assertFalse(obj.is_blocked)
        self.assertTrue(obj.is_watched)
        self.assertEqual(obj.standing, -10)
        self.assertEqual({x.label_id for x in obj.labels.all()}, {2})

    def test_should_not_update_when_data_has_not_changed(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(character=self.character, label_id=1)
        create_character_contact_label(character=self.character, label_id=2)

        self.character.update_contacts()
        obj = self.character.contacts.get(eve_entity_id=1101)
        obj.is_watched = False
        obj.save()

        # when
        result = self.character.update_contacts()

        # then
        self.assertFalse(result.is_changed)
        obj = self.character.contacts.get(eve_entity_id=1101)
        self.assertFalse(obj.is_watched)

    def test_should_always_update_when_data_has_not_changed_and_forced(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contact_label(character=self.character, label_id=1)
        create_character_contact_label(character=self.character, label_id=2)

        self.character.update_contacts()
        obj = self.character.contacts.get(eve_entity_id=1101)
        obj.is_watched = False
        obj.save()

        # when
        result = self.character.update_contacts(force_update=True)

        # then
        self.assertFalse(result.is_changed)
        self.assertTrue(result.is_updated)
        obj = self.character.contacts.get(eve_entity_id=1101)
        self.assertTrue(obj.is_watched)


@patch(MODULE_PATH + ".esi")
class TestCharacterContractsUpdate(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        load_locations()
        cls.character = create_memberaudit_character(1001)
        cls.jita_44 = Location.objects.get(id=60003760)
        cls.structure_1 = Location.objects.get(id=1000000000001)

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_can_create_new_courier_contract(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterContract.objects.update_or_create_esi(self.character)
        # then
        self.assertSetEqual(
            set(self.character.contracts.values_list("contract_id", flat=True)),
            {100000001, 100000002, 100000003},
        )

        obj = self.character.contracts.get(contract_id=100000001)
        self.assertEqual(obj.contract_type, CharacterContract.TYPE_COURIER)
        self.assertEqual(obj.acceptor, EveEntity.objects.get(id=1101))
        self.assertEqual(obj.assignee, EveEntity.objects.get(id=2101))
        self.assertEqual(obj.availability, CharacterContract.AVAILABILITY_PERSONAL)
        self.assertIsNone(obj.buyout)
        self.assertEqual(float(obj.collateral), 550000000.0)
        self.assertEqual(obj.date_accepted, parse_datetime("2019-10-06T13:15:21Z"))
        self.assertEqual(obj.date_completed, parse_datetime("2019-10-07T13:15:21Z"))
        self.assertEqual(obj.date_expired, parse_datetime("2019-10-09T13:15:21Z"))
        self.assertEqual(obj.date_issued, parse_datetime("2019-10-02T13:15:21Z"))
        self.assertEqual(obj.days_to_complete, 3)
        self.assertEqual(obj.end_location, self.structure_1)
        self.assertFalse(obj.for_corporation)
        self.assertEqual(obj.issuer_corporation, EveEntity.objects.get(id=2001))
        self.assertEqual(obj.issuer, EveEntity.objects.get(id=1001))
        self.assertEqual(float(obj.price), 0.0)
        self.assertEqual(float(obj.reward), 500000000.0)
        self.assertEqual(obj.start_location, self.jita_44)
        self.assertEqual(obj.status, CharacterContract.STATUS_IN_PROGRESS)
        self.assertEqual(obj.title, "Test 1")
        self.assertEqual(obj.volume, 486000.0)

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_should_keep_old_contracts_when_updating(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contract(
            character=self.character,
            contract_id=190000001,
            contract_type=CharacterContract.TYPE_COURIER,
            assignee=EveEntity.objects.get(id=1002),
            issuer=EveEntity.objects.get(id=1001),
            issuer_corporation=EveEntity.objects.get(id=2001),
            status=CharacterContract.STATUS_IN_PROGRESS,
            start_location=self.jita_44,
            end_location=self.structure_1,
        )
        # when
        CharacterContract.objects.update_or_create_esi(self.character)
        # then
        self.assertEqual(self.character.contracts.count(), 4)

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_should_update_existing_contracts(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contract(
            character=self.character,
            contract_id=100000001,
            availability=CharacterContract.AVAILABILITY_PERSONAL,
            contract_type=CharacterContract.TYPE_COURIER,
            assignee=EveEntity.objects.get(id=2101),
            date_issued=parse_datetime("2019-10-02T13:15:21Z"),
            date_expired=parse_datetime("2019-10-09T13:15:21Z"),
            for_corporation=False,
            issuer=EveEntity.objects.get(id=1001),
            issuer_corporation=EveEntity.objects.get(id=2001),
            status=CharacterContract.STATUS_OUTSTANDING,
            start_location=self.jita_44,
            end_location=self.structure_1,
            title="Test 1",
            collateral=550000000,
            reward=500000000,
            volume=486000,
            days_to_complete=3,
        )
        # when
        CharacterContract.objects.update_or_create_esi(self.character)
        # then
        obj = self.character.contracts.get(contract_id=100000001)
        self.assertEqual(obj.contract_type, CharacterContract.TYPE_COURIER)
        self.assertEqual(obj.acceptor, EveEntity.objects.get(id=1101))
        self.assertEqual(obj.assignee, EveEntity.objects.get(id=2101))
        self.assertEqual(obj.availability, CharacterContract.AVAILABILITY_PERSONAL)
        self.assertIsNone(obj.buyout)
        self.assertEqual(float(obj.collateral), 550000000.0)
        self.assertEqual(obj.date_accepted, parse_datetime("2019-10-06T13:15:21Z"))
        self.assertEqual(obj.date_completed, parse_datetime("2019-10-07T13:15:21Z"))
        self.assertEqual(obj.date_expired, parse_datetime("2019-10-09T13:15:21Z"))
        self.assertEqual(obj.date_issued, parse_datetime("2019-10-02T13:15:21Z"))
        self.assertEqual(obj.days_to_complete, 3)
        self.assertEqual(obj.end_location, self.structure_1)
        self.assertFalse(obj.for_corporation)
        self.assertEqual(obj.issuer_corporation, EveEntity.objects.get(id=2001))
        self.assertEqual(obj.issuer, EveEntity.objects.get(id=1001))
        self.assertEqual(float(obj.reward), 500000000.0)
        self.assertEqual(obj.start_location, self.jita_44)
        self.assertEqual(obj.status, CharacterContract.STATUS_IN_PROGRESS)
        self.assertEqual(obj.title, "Test 1")
        self.assertEqual(obj.volume, 486000.0)

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_should_skip_updates_when_there_is_no_change(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        self.character.update_contract_headers()
        obj = self.character.contracts.get(contract_id=100000001)
        obj.status = CharacterContract.STATUS_FINISHED
        obj.save()
        # when
        CharacterContract.objects.update_or_create_esi(self.character)
        # then
        obj = self.character.contracts.get(contract_id=100000001)
        self.assertEqual(obj.status, CharacterContract.STATUS_FINISHED)

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_always_update_when_forced(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        self.character.update_contract_headers()
        obj = self.character.contracts.get(contract_id=100000001)
        obj.status = CharacterContract.STATUS_FINISHED
        obj.save()
        # when
        CharacterContract.objects.update_or_create_esi(
            self.character, force_update=True
        )
        # then
        obj = self.character.contracts.get(contract_id=100000001)
        self.assertEqual(obj.status, CharacterContract.STATUS_IN_PROGRESS)

    @patch(
        MODULE_PATH + ".data_retention_cutoff",
        lambda: dt.datetime(2019, 10, 11, 1, 15, tzinfo=dt.timezone.utc),
    )
    def test_when_updating_then_use_retention_limit(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        # when
        CharacterContract.objects.update_or_create_esi(self.character)
        # then
        self.assertSetEqual(
            set(self.character.contracts.values_list("contract_id", flat=True)),
            {100000002, 100000003},
        )

    @patch(
        MODULE_PATH + ".data_retention_cutoff",
        lambda: dt.datetime(2019, 10, 6, 1, 15, tzinfo=dt.timezone.utc),
    )
    def test_when_retention_limit_is_set_then_remove_outdated_contracts(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        create_character_contract(
            character=self.character,
            contract_id=100_000_004,
            availability=CharacterContract.AVAILABILITY_PERSONAL,
            contract_type=CharacterContract.TYPE_COURIER,
            assignee=EveEntity.objects.get(id=2101),
            date_issued=parse_datetime("2019-09-02T13:15:21Z"),
            date_expired=parse_datetime("2019-09-09T13:15:21Z"),
            for_corporation=False,
            issuer=EveEntity.objects.get(id=1001),
            issuer_corporation=EveEntity.objects.get(id=2001),
            status=CharacterContract.STATUS_OUTSTANDING,
            start_location=self.jita_44,
            end_location=self.structure_1,
            title="This contract is too old",
            collateral=550000000,
            reward=500000000,
            volume=486000,
            days_to_complete=3,
        )
        # when
        CharacterContract.objects.update_or_create_esi(self.character)
        # then
        self.assertSetEqual(
            set(self.character.contracts.values_list("contract_id", flat=True)),
            {100000001, 100000002, 100000003},
        )

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_can_create_new_item_exchange_contract(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        CharacterContract.objects.update_or_create_esi(self.character)
        contract = self.character.contracts.get(contract_id=100000002)
        self.assertEqual(contract.contract_type, CharacterContract.TYPE_ITEM_EXCHANGE)
        self.assertEqual(float(contract.price), 270000000.0)
        self.assertEqual(contract.volume, 486000.0)
        self.assertEqual(contract.status, CharacterContract.STATUS_FINISHED)
        # when
        CharacterContractItem.objects.update_or_create_esi(self.character, contract)
        # then
        self.assertEqual(contract.items.count(), 2)

        item = contract.items.get(record_id=1)
        self.assertTrue(item.is_included)
        self.assertFalse(item.is_singleton)
        self.assertEqual(item.quantity, 3)
        self.assertEqual(item.eve_type, EveType.objects.get(id=19540))

        item = contract.items.get(record_id=2)
        self.assertTrue(item.is_included)
        self.assertFalse(item.is_singleton)
        self.assertEqual(item.quantity, 5)
        self.assertEqual(item.raw_quantity, -1)
        self.assertEqual(item.eve_type, EveType.objects.get(id=19551))

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_can_create_auction_contract(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        CharacterContract.objects.update_or_create_esi(self.character)
        contract = self.character.contracts.get(contract_id=100000003)
        self.assertEqual(contract.contract_type, CharacterContract.TYPE_AUCTION)
        self.assertEqual(float(contract.buyout), 200_000_000.0)
        self.assertEqual(float(contract.price), 20_000_000.0)
        self.assertEqual(contract.volume, 400.0)
        self.assertEqual(contract.status, CharacterContract.STATUS_OUTSTANDING)
        CharacterContractItem.objects.update_or_create_esi(self.character, contract)
        self.assertEqual(contract.items.count(), 1)
        item = contract.items.get(record_id=1)
        self.assertTrue(item.is_included)
        self.assertFalse(item.is_singleton)
        self.assertEqual(item.quantity, 3)
        self.assertEqual(item.eve_type, EveType.objects.get(id=19540))
        # when
        CharacterContractBid.objects.update_or_create_esi(self.character, contract)
        # then
        self.assertEqual(contract.bids.count(), 1)
        bid = contract.bids.get(bid_id=1)
        self.assertEqual(float(bid.amount), 1_000_000.23)
        self.assertEqual(bid.date_bid, parse_datetime("2017-01-01T10:10:10Z"))
        self.assertEqual(bid.bidder, EveEntity.objects.get(id=1101))

    @patch(MODULE_PATH + ".data_retention_cutoff", lambda: None)
    def test_can_add_new_bids_to_auction_contract(self, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        contract = create_character_contract(
            character=self.character,
            contract_id=100000003,
            availability=CharacterContract.AVAILABILITY_PERSONAL,
            contract_type=CharacterContract.TYPE_AUCTION,
            assignee=EveEntity.objects.get(id=2101),
            date_issued=parse_datetime("2019-10-02T13:15:21Z"),
            date_expired=parse_datetime("2019-10-09T13:15:21Z"),
            for_corporation=False,
            issuer=EveEntity.objects.get(id=1001),
            issuer_corporation=EveEntity.objects.get(id=2001),
            status=CharacterContract.STATUS_OUTSTANDING,
            start_location=self.jita_44,
            end_location=self.jita_44,
            buyout=200_000_000,
            price=20_000_000,
            volume=400,
        )
        create_character_contract_bid(
            contract=contract,
            bid_id=2,
            amount=21_000_000,
            bidder=EveEntity.objects.get(id=1003),
            date_bid=now(),
        )
        self.character.update_contract_headers()
        # when
        self.character.update_contract_bids(contract=contract)
        # then
        contract.refresh_from_db()
        self.assertEqual(contract.bids.count(), 2)

        bid = contract.bids.get(bid_id=1)
        self.assertEqual(float(bid.amount), 1_000_000.23)
        self.assertEqual(bid.date_bid, parse_datetime("2017-01-01T10:10:10Z"))
        self.assertEqual(bid.bidder, EveEntity.objects.get(id=1101))

        bid = contract.bids.get(bid_id=2)
        self.assertEqual(float(bid.amount), 21_000_000)


class TestCharacterContractBidManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities()
        cls.character = create_memberaudit_character(1001)

    def test_should_do_nothing_when_there_are_no_bids(self):
        # given
        contract = create_character_contract(
            character=self.character, contract_type=CharacterContract.TYPE_AUCTION
        )
        # when
        CharacterContractBid.objects._update_or_create_objs(
            contract=contract, bids_list={}
        )
        # then
        self.assertEqual(CharacterContractBid.objects.count(), 0)

    def test_should_do_nothing_when_there_are_no_new_bids(self):
        # given
        contract = create_character_contract(
            character=self.character, contract_type=CharacterContract.TYPE_AUCTION
        )
        bidder = EveEntity.objects.get(id=1002)
        bid = create_character_contract_bid(contract=contract, bidder=bidder)
        bids_list = {
            bid.bid_id: {
                "amount": bid.amount,
                "bid_id": bid.bid_id,
                "bidder_id": bidder.id,
                "date_bid": bid.date_bid,
            }
        }
        # when
        CharacterContractBid.objects._update_or_create_objs(
            contract=contract, bids_list=bids_list
        )
        # then
        self.assertEqual(CharacterContractBid.objects.count(), 1)
